/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/09/26
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy;

import java.util.Arrays;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.Validate;

import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.facade.JiemamyFacade;
import org.jiemamy.internal.ArtemisJiemamyFactory;
import org.jiemamy.internal.EventBrokerImpl;
import org.jiemamy.internal.ReferenceResolverImpl;
import org.jiemamy.serializer.JiemamySerializer;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.utils.enhancer.Enhance;
import org.jiemamy.utils.enhancer.EnhanceException;
import org.jiemamy.utils.enhancer.FactoryEnhancer;
import org.jiemamy.xml.CoreNamespace;
import org.jiemamy.xml.JiemamyNamespace;

/**
 * Jiemamy仕様の参照実装(コードネーム:Artemis)を表す型。
 * 
 * <p>このクラスのコンストラクタ引数に何も与えずにインスタンス生成した場合、Jiemamy Core仕様のみをサポートする
 * 実装として機能する。引数に{@link ArtemisExtender}型のインスタンスを与えることにより、実装がサポートする
 * 範囲を拡張することができる。</p>
 * 
 * <p>Jiemamy Projectとして提供する{@link ArtemisExtender}は、現在(v0.2.0)のところ
 * jiemamy-view-(version).jarに含まれるorg.jiemamy.ArtemisView及び、jiemamy-event-(version).jarに含まれる
 * org.jiemamy.ArtemisEventの2種類である。</p>
 * 
 * <p>以上の{@link ArtemisExtender}はオフィシャルな仕様であるが、ユーザは{@link ArtemisExtender}を独自に実装して
 * Jiemamyを拡張することも可能である。</p>
 * 
 * @author daisuke
 */
public class Artemis implements JiemamyImplementation {
	
	/** 参照オブジェクトから実体モデルを解決するリゾルバ */
	private final ReferenceResolver resolver = new ReferenceResolverImpl();
	
	/** イベント通知を適切に行うブローカー */
	private final EventBroker eventBroker = new EventBrokerImpl();
	
	/** 実装に施された拡張の配列 */
	private final ArtemisExtender[] extenders;
	
	/** 拡張をほどこされたファクトリの{@link Class} */
	private Class<? extends JiemamyFactory> enhancedFactoryClass;
	
	/**
	 * 生成したファクトリのキャッシュ
	 * 
	 * 1つの {@link Jiemamy}インスタンスにつき、{@link JiemamyFactory}は1つでなければならない。
	 */
	private Map<Jiemamy, JiemamyFactory> factoryCache = new IdentityHashMap<Jiemamy, JiemamyFactory>();
	

	/**
	 * インスタンスを生成する。
	 */
	public Artemis() {
		this(new ArtemisExtender[0]);
	}
	
	/**
	 * インスタンスを生成する。
	 * 
	 * @param extenders Artemis拡張オブジェクト
	 */
	public Artemis(ArtemisExtender... extenders) {
		this.extenders = extenders;
		
		List<Enhance> allEnhances = CollectionsUtil.newArrayList();
		for (ArtemisExtender extender : extenders) {
			allEnhances.addAll(extender.getEnhances());
		}
		
		FactoryEnhancer<JiemamyFactory> enhancer =
				new FactoryEnhancer<JiemamyFactory>(JiemamyFactory.class, ArtemisJiemamyFactory.class, allEnhances);
		
		try {
			enhancedFactoryClass = enhancer.getEnhanced();
		} catch (EnhanceException e) {
			throw new UnexpectedConditionError("enhancerが要件を満たしていない可能性。", e);
		}
		assert enhancedFactoryClass != null;
	}
	
	public EventBroker getEventBroker() {
		assert eventBroker != null;
		return eventBroker;
	}
	
	public JiemamyFactory getFactory(Jiemamy jiemamy) {
		Validate.notNull(jiemamy);
		JiemamyFactory factory = factoryCache.get(jiemamy);
		if (factory == null) {
			try {
				factory = enhancedFactoryClass.newInstance();
				factory.setImplementation(this);
				factory.setJiemamy(jiemamy);
				factoryCache.put(jiemamy, factory);
			} catch (InstantiationException e) {
				throw new UnexpectedConditionError("JiemamyFactoryのインスタンス化に失敗。実装ミス？", e);
			} catch (IllegalAccessException e) {
				throw new UnexpectedConditionError("JiemamyFactoryのインスタンス化に失敗。実装ミス？", e);
			}
		}
		return factory;
	}
	
	public JiemamyNamespace[] getNamespaces() {
		List<JiemamyNamespace> namespaces = CollectionsUtil.newArrayList();
		namespaces.addAll(Arrays.asList(CoreNamespace.values()));
		
		for (ArtemisExtender extender : extenders) {
			namespaces.addAll(Arrays.asList(extender.getNamespaces()));
		}
		
		return namespaces.toArray(new JiemamyNamespace[namespaces.size()]);
	}
	
	public ReferenceResolver getReferenceResolver() {
		assert resolver != null;
		return resolver;
	}
	
	public JiemamySerializer getSerializer(Jiemamy jiemamy) {
		return getFactory(jiemamy).getSerializer();
	}
	
	public Set<OptionalSpec> getSupportedSpecs() {
		Set<OptionalSpec> supportedSpecs = CollectionsUtil.newHashSet();
		for (ArtemisExtender enhance : extenders) {
			supportedSpecs.addAll(enhance.getSpecs());
		}
		return supportedSpecs;
	}
	
	public Version getVersion() {
		return ArtemisVersion.INSTANCE;
	}
	
	public <T extends JiemamyFacade>T newFacade(Jiemamy jiemamy, Class<T> clazz) {
		return getFactory(jiemamy).newFacade(clazz);
	}
	

	private static class ArtemisVersion implements Version {
		
		static final ArtemisVersion INSTANCE = new ArtemisVersion();
		
		private final int release = 0;
		

		private ArtemisVersion() {
		}
		
		public int getMajor() {
			return MAJOR;
		}
		
		public int getMinor() {
			return MINOR;
		}
		
		public int getRelease() {
			return release;
		}
		
		public boolean isSnapshot() {
			return SNAPSHOT;
		}
		
		@Override
		public String toString() {
			StringBuilder sb = new StringBuilder();
			sb.append(MAJOR);
			sb.append(".");
			sb.append(MINOR);
			sb.append(".");
			sb.append(release);
			if (SNAPSHOT) {
				sb.append("-SNAPSHOT");
			}
			return sb.toString();
		}
	}
}
