/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/12/29
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy;

import java.util.Set;
import java.util.UUID;

import org.jiemamy.exception.IllegalImplementationException;
import org.jiemamy.experimental.builder.JiemamyBuilder;
import org.jiemamy.facade.JiemamyFacade;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.datatype.DataType;
import org.jiemamy.model.datatype.DataTypeMold;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.datatype.DomainRef;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.TableRef;
import org.jiemamy.serializer.JiemamySerializer;

/**
 * Jiemamy関連のインスタンスを生成するファクトリインターフェイス。
 * 
 * <p>ユーザは、このインターフェイスを介して、各種Jiemamy関連インスタンスを取得する。</p>
 * 
 * @since 0.2
 * @author daisuke
 */
public interface JiemamyFactory {
	
	/**
	* Jiemamy実装を取得する。
	* 
	* <p>{@code null}を返してはならない。</p>
	* 
	* @return Jiemamy実装
	* @since 0.2
	*/
	JiemamyImplementation getImplementation();
	
	/**
	 * インスタンス空間に属する唯一の{@link RootModel}を取得する。存在しない場合は生成する。
	 * 
	 * <p>{@code null}を返してはならない。</p>
	 * 
	 * @return {@link RootModel}
	 * @since 0.2
	 */
	RootModel getRootModel();
	
	/**
	 * 指定したIDを持つ、このインスタンス空間に属する唯一の{@link RootModel}を取得する。存在しない場合は生成する。
	 * 
	 * <p>{@code null}を返してはならない。</p>
	 * 
	 * @param id モデルID
	 * @return {@link RootModel}
	 * @throws IllegalStateException インスタンス空間に既に別のIDを持つ{@link RootModel}が存在した場合
	 * @since 0.2
	 */
	RootModel getRootModel(UUID id);
	
	/**
	 * シリアライザを取得する。
	 * 
	 * <p>{@code null}を返してはならない。</p>
	 * 
	 * @return シリアライザ
	 * @throws UnsupportedOperationException 実装がモデルのシリアライズをサポートしない場合
	 * @since 0.2
	 */
	JiemamySerializer getSerializer();
	
	/**
	 * 追加仕様のサポート状態を取得する。
	 * 
	 * <p>{@code null}を返してはならない。</p>
	 * 
	 * @return サポートされている追加仕様の{@link Set}
	 * @see JiemamyImplementation#getSupportedSpecs()
	 * @since 0.2
	 */
	Set<OptionalSpec> getSupportedSpecs();
	
	/**
	 * インスタンス空間に属する唯一の{@link RootModel}が生成済みであるかどうか調べる。
	 * 
	 * @return 生成済みの場合は{@code true}、そうでない場合は{@code false}
	 * @since 0.2
	 */
	boolean hasRootModel();
	
	/**
	 * 新しいアダプタインスタンスを生成する。
	 * 
	 * @param <T> 生成するアダプタの種類を表す型
	 * @param clazz 生成するアダプタの種類を表す型
	 * @return 生成されたアダプタ
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	<T>T newAdapter(Class<T> clazz);
	
	/**
	 * 新しいモデルビルダを生成する。
	 * 
	 * @param <M> モデルの型
	 * @param <B> モデルビルダの型
	 * @param clazz 生成したいモデルの型
	 * @return モデルビルダ
	 * @throws UnsupportedOperationException 実装がビルダをサポートしない場合
	 * @since 0.2
	 */
	<M extends JiemamyElement, B extends JiemamyBuilder<M>>B newBuilder(Class<B> clazz);
	
	/**
	 * 鋳型情報から、新しい{@link DataType}型のインスタンスを生成する。
	 * 
	 * @param <T> データ型の型
	 * @param mold 鋳型情報
	 * @return {@link DataType}型のインスタンス
	 * @throws IllegalImplementationException moldに不正なアダプタが登録されている場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	<T extends DataType>T newDataType(DataTypeMold<T> mold);
	
	/**
	 * 新しいファサードのインスタンスを生成し、取得する。
	 * 
	 * @param <T> 取得するファサードの型
	 * @param clazz 取得するファサードの型
	 * @return 新しいファサード
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	<T extends JiemamyFacade>T newFacade(Class<T> clazz);
	
	/**
	 * 新しいモデルを生成する。
	 * 
	 * <p>{@code null}を返してはならない。</p>
	 * 
	 * @param <T> 生成するモデルの種類を表す型
	 * @param clazz 生成するモデルの種類を表す型
	 * @return 生成されたモデル
	 * @throws IllegalStateException ファクトリが初期化されていない場合
	 * @throws IllegalArgumentException 引数clazzに責務外のクラスを指定して、インスタンスを取得しようとした場合
	 * @throws IllegalArgumentException 引数として{@code null}を与えた場合
	 * @since 0.2
	 */
	<T extends JiemamyElement>T newModel(Class<T> clazz);
	
	/**
	 * 指定したIDを持つ新しいモデルを生成する。
	 * 
	 * <p>{@code null}を返してはならない。</p>
	 * 
	 * @param <T> 生成するモデルの種類を表す型
	 * @param clazz 生成するモデルの種類を表す型
	 * @param id モデルID
	 * @return 生成されたモデル
	 * @throws IllegalStateException ファクトリが初期化されていない場合
	 * @throws IllegalArgumentException 引数clazzに責務外のクラスを指定して、インスタンスを取得しようとした場合
	 * @throws IllegalArgumentException 引数として{@code null}を与えた場合
	 * @since 0.2
	 */
	<T extends JiemamyElement>T newModel(Class<T> clazz, UUID id);
	
	/**
	 * {@link ColumnModel}の参照オブジェクトを生成する。
	 * 
	 * <p>{@code null}を返してはならない。</p>
	 * 
	 * @param columnModel 対象の{@link ColumnModel}
	 * @return {@link ColumnModel}の参照オブジェクト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	ColumnRef newReference(ColumnModel columnModel);
	
	/**
	 * {@link DomainModel}の参照オブジェクトを生成する。
	 * 
	 * <p>{@code null}を返してはならない。</p>
	 * 
	 * @param domainModel 対象の{@link DomainModel}
	 * @return {@link DomainModel}の参照オブジェクト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	DomainRef newReference(DomainModel domainModel);
	
//	/**
//	 * 新しい{@link RootModel}を生成する。
//	 * 
//	 * <p>{@code null}を返してはならない。</p>
//	 * 
//	 * @return 新しい{@link RootModel}
//	 */
//	RootModel newRootModel();
//	
//	/**
//	 * 指定したIDを持つ新しい{@link RootModel}を生成する。
//	 * 
//	 * <p>{@code null}を返してはならない。</p>
//	 * 
//	 * @param id モデルID
//	 * @return 新しい{@link RootModel}
//	 */
//	RootModel newRootModel(UUID id);
	
	/**
	 * {@link TableModel}の参照オブジェクトを生成する。
	 * 
	 * <p>{@code null}を返してはならない。</p>
	 * 
	 * @param tableModel 対象の{@link TableModel}
	 * @return {@link TableModel}の参照オブジェクト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	TableRef newReference(TableModel tableModel);
	
	/**
	 * ファクトリが使用するJiemamy実装を設定する。
	 * 
	 * <p>このメソッドは、ファクトリのインスタンス生成直後に一度だけ呼ばれるべきである。
	 * APIユーザによる直接の呼び出しは意図されていない。</p>
	 * 
	 * @param implementation Jiemamy実装
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws IllegalStateException 以前に既にこのメソッドが一度呼ばれている場合
	 * @since 0.2
	 */
	void setImplementation(JiemamyImplementation implementation);
	
	/**
	 * ファクトリが生成するモデルオブジェクトが属する「インスタンス空間」を設定する。
	 * 
	 * <p>このメソッドは、ファクトリのインスタンス生成直後に一度だけ呼ばれるべきである。
	 * APIユーザによる直接の呼び出しは意図されていない。</p>
	 * 
	 * @param jiemamy インスタンス空間
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws IllegalStateException 以前に既にこのメソッドが一度呼ばれている場合
	 * @since 0.2
	 */
	void setJiemamy(Jiemamy jiemamy);
	
}
