/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/10/13
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy;

import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.constraint.CheckConstraint;
import org.jiemamy.model.attribute.constraint.ColumnCheckConstraint;
import org.jiemamy.model.attribute.constraint.ConstraintModel;
import org.jiemamy.model.attribute.constraint.Deferrability;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.KeyConstraint;
import org.jiemamy.model.attribute.constraint.NotNullConstraint;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.TableCheckConstraint;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.attribute.constraint.ValueConstraint;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.dataset.RecordModel;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.ViewModel;
import org.jiemamy.model.index.IndexColumnModel;
import org.jiemamy.model.index.IndexModel;

/**
 * 各モデルの「概念的な」（実装としては強要しない）プロパティを表すインターフェイス。
 * 
 * @param <T> モデルの型
 * @since 0.2
 * @author daisuke
 */
public interface JiemamyProperty<T extends JiemamyElement> {
	
	/**
	 * プロパティ名を返す。
	 * 
	 * @return プロパティ名
	 * @see Enum#name()
	 * @since 0.2
	 */
	String name();
	

	/**
	 * {@link AttributeModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum AttributeProperty implements JiemamyProperty<AttributeModel> {
		
		/** {@link AttributeModel#getName()}で取得できるプロパティ */
		name,

		/** {@link AttributeModel#getLogicalName()}で取得できるプロパティ */
		logicalName,

		/** {@link AttributeModel#getDescription()}で取得できるプロパティ */
		description
	}
	
	/**
	 * {@link CheckConstraint}のプロパティを表す列挙型。
	 * 
	 * @author daisuke
	 */
	public enum CheckConstraintProperty implements JiemamyProperty<CheckConstraint> {
		
		/** {@link CheckConstraint#getExpression()}で取得できるプロパティ */
		expression
	}
	
	/**
	 * {@link ColumnCheckConstraint}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum ColumnCheckModelProperty implements JiemamyProperty<ColumnCheckConstraint> {
		
	}
	
	/**
	 * {@link ColumnModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum ColumnProperty implements JiemamyProperty<ColumnModel> {
		
		/** {@link ColumnModel#getDataType()}で取得できるプロパティ */
		dataType,

		/** {@link ColumnModel#getDefaultValue()}で取得できるプロパティ */
		defaultValue,

		/** {@link ColumnModel#getNotNullConstraint()}で取得できるプロパティ */
		notNullConstraint,

		/** {@link ColumnModel#getUniqueKey()}で取得できるプロパティ */
		uniqueKey,

		/** {@link ColumnModel#getPrimaryKey()}で取得できるプロパティ */
		primaryKey,

		/** {@link ColumnModel#getCheckConstraint()}で取得できるプロパティ */
		checkConstraint
	}
	
	/**
	 * {@link ConstraintModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum ConstraintProperty implements JiemamyProperty<ConstraintModel> {
		
		/** {@link ConstraintModel#getName()}で取得できるプロパティ */
		name
	}
	
	/**
	 * {@link DataSetModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum DataSetProperty implements JiemamyProperty<DataSetModel> {
		
		/** {@link DataSetModel#getName()}で取得できるプロパティ */
		name,

		/** {@link DataSetModel#getRecords()}で取得できるプロパティ */
		records
	}
	
	/**
	 * {@link Deferrability}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum DeferrabilityProperty implements JiemamyProperty<Deferrability> {
		
		/** {@link Deferrability#isDeferrable()}で取得できるプロパティ */
		deferrable,

		/** {@link Deferrability#getInitiallyCheckTime()}で取得できるプロパティ */
		initiallyCheckTime
	}
	
	/**
	 * {@link DomainModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum DomainProperty implements JiemamyProperty<DomainModel> {
		
		/** {@link DomainModel#getName()}で取得できるプロパティ */
		name,

		/** {@link DomainModel#getLogicalName()}で取得できるプロパティ */
		logicalName,

		/** {@link DomainModel#getDescription()}で取得できるプロパティ */
		description,

		/** {@link DomainModel#getDataType()}で取得できるプロパティ */
		dataType,

		/** {@link DomainModel#getNotNullConstraint()}で取得できるプロパティ */
		notNullConstraint,

		/** {@link DomainModel#getCheckConstraint()}で取得できるプロパティ */
		checkConstraint
	}
	
	/**
	 * {@link EntityModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum EntityProperty implements JiemamyProperty<EntityModel> {
		/** {@link EntityModel#getName()}で取得できるプロパティ */
		name,

		/** {@link EntityModel#getLogicalName()}で取得できるプロパティ */
		logicalName,

		/** {@link EntityModel#getDescription()}で取得できるプロパティ */
		description,

		/** {@link EntityModel#getBeginScript()}で取得できるプロパティ */
		beginScript,

		/** {@link EntityModel#getEndScript()}で取得できるプロパティ */
		endScript
	}
	
	/**
	 * {@link ForeignKey}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum ForeignKeyProperty implements JiemamyProperty<ForeignKey> {
		
		/** {@link ForeignKey#getReferenceColumns()}で取得できるプロパティ */
		referenceColumns,

		/** {@link ForeignKey#getMatchType()}で取得できるプロパティ */
		matchType,

		/** {@link ForeignKey#getOnDelete()}で取得できるプロパティ */
		onDelete,

		/** {@link ForeignKey#getOnUpdate()}で取得できるプロパティ */
		onUpdate
	}
	
	/**
	 * {@link IndexColumnModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum IndexColumnProperty implements JiemamyProperty<IndexColumnModel> {
		
		/** {@link IndexColumnModel#getColumnRef()}で取得できるプロパティ */
		columnRef,

		/** {@link IndexColumnModel#getSortOrder()}で取得できるプロパティ */
		sortOrder
	}
	
	/**
	 * {@link IndexModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum IndexProperty implements JiemamyProperty<IndexModel> {
		
		/** {@link IndexModel#getName()}で取得できるプロパティ */
		name,

		/** {@link IndexModel#getIndexColumns()}で取得できるプロパティ */
		indexColumns,

		/** {@link IndexModel#isUnique()}で取得できるプロパティ */
		unique
	}
	
	/**
	 * {@link KeyConstraint}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum KeyConstraintModelProperty implements JiemamyProperty<KeyConstraint> {
		
		/** {@link KeyConstraint#getDeferrability()}で取得できるプロパティ */
		deferrability,

		/** {@link KeyConstraint#getKeyColumns()}で取得できるプロパティ */
		keyColumns
	}
	
	/**
	 * {@link NotNullConstraint}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum NotNullConstraintProperty implements JiemamyProperty<NotNullConstraint> {
		
	}
	
	/**
	 * {@link PrimaryKey}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum PrimaryKeyProperty implements JiemamyProperty<PrimaryKey> {
		
	}
	
	/**
	 * {@link RecordModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum RecordProperty implements JiemamyProperty<RecordModel> {
		
		/** {@link RecordModel#getValues()}で取得できるプロパティ */
		values
	}
	
	/**
	 * {@link RootModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum RootProperty implements JiemamyProperty<RootModel> {
		
		/** {@link RootModel#getDialectClassName()}で取得できるプロパティ */
		dialectClassName,

		/** {@link RootModel#getBeginScript()}で取得できるプロパティ */
		beginScript,

		/** {@link RootModel#getEndScript()}で取得できるプロパティ */
		endScript,

		/** {@link RootModel#getDescription()}で取得できるプロパティ */
		description,

		/** {@link RootModel#getSchemaName()}で取得できるプロパティ */
		schemaName,

		/** {@link RootModel#getDomains()}で取得できるプロパティ */
		domains,

		/** {@link RootModel#getEntities()}で取得できるプロパティ */
		entities,

		/** {@link RootModel#getDataSets()}で取得できるプロパティ */
		dataSets
	}
	
	/**
	 * {@link ColumnCheckConstraint}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum TableCheckModelProperty implements JiemamyProperty<TableCheckConstraint> {
		
	}
	
	/**
	 * {@link TableModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum TableProperty implements JiemamyProperty<TableModel> {
		
		/** {@link TableModel#getAttributes()}で取得できるプロパティ */
		attributes,

		/** {@link TableModel#getIndexes()}で取得できるプロパティ */
		indexes
	}
	
	/**
	 * {@link UniqueKey}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum UniqueKeyProperty implements JiemamyProperty<UniqueKey> {
		
	}
	
	/**
	 * {@link ValueConstraint}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum ValueConstraintProperty implements JiemamyProperty<ValueConstraint> {
		
	}
	
	/**
	 * {@link ViewModel}のプロパティを表す列挙型。
	 * 
	 * @since 0.2
	 * @author daisuke
	 */
	public enum ViewProperty implements JiemamyProperty<ViewModel> {
		
		/** {@link ViewModel#getDefinition()}で取得できるプロパティ */
		definition
	}
	
}
