/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/07/07
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.dialect.internal;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.Validate;

import org.jiemamy.dialect.BuiltinDataTypeMold;
import org.jiemamy.dialect.ConnectionAdvisor;
import org.jiemamy.dialect.DataTypeResolver;
import org.jiemamy.dialect.Dialect;
import org.jiemamy.dialect.EmitConfig;
import org.jiemamy.dialect.ImportMetadataConfig;
import org.jiemamy.dialect.MoldManager;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.datatype.BuiltinDataType;
import org.jiemamy.model.datatype.DataType;
import org.jiemamy.model.datatype.DataTypeCategory;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.sql.SqlStatement;
import org.jiemamy.utils.metadata.TypeSafeDatabaseMetaData;
import org.jiemamy.validator.AllValidator;
import org.jiemamy.validator.Validator;

/**
 * SQL方言実装クラスの標準的な実装クラス。
 * 
 * @author daisuke
 */
public abstract class AbstractDialect implements Dialect {
	
	/** データ型（{@link DataType}）を中心とした各種解決を行うロジック */
	private final DataTypeResolver dataTypeResolver;
	
	/** DBメタデータの読み出し前後に実行するロジック */
	private final ConnectionAdvisor advisor;
	
	/** SQLエミッタ */
	private final SqlEmitter emitter;
	
	/** データ型鋳型マネージャ */
	private final MoldManager moldManager;
	
	/** DBからエンティティ情報をインポートするビジター */
	private final EntityImportVisitor entityImportVisitor;
	
	/** DBから外部キー情報をインポートするビジター */
	private final ForeignKeyImportVisitor foreignKeyImportVisitor;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param initializer SQL方言の各ストラテジを提供するイニシャライザ
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws IllegalArgumentException initializerから取得される各ストラテジの何れかが{@code null}の場合
	 */
	public AbstractDialect(DialectInitializer initializer) {
		Validate.notNull(initializer);
		Validate.notNull(initializer.getMoldManager());
		Validate.notNull(initializer.getDataTypeResolver());
		Validate.notNull(initializer.getConnectionAdvisor());
		Validate.notNull(initializer.getSqlEmitter());
		Validate.notNull(initializer.getEntityImportVisitor());
		Validate.notNull(initializer.getForeignKeyImportVisitor());
		moldManager = initializer.getMoldManager();
		dataTypeResolver = initializer.getDataTypeResolver();
		advisor = initializer.getConnectionAdvisor();
		emitter = initializer.getSqlEmitter();
		entityImportVisitor = initializer.getEntityImportVisitor();
		foreignKeyImportVisitor = initializer.getForeignKeyImportVisitor();
	}
	
	public List<SqlStatement> emitStatements(RootModel rootModel, EmitConfig config) {
		return emitter.emit(rootModel, config);
	}
	
	public BuiltinDataTypeMold findDataTypeMold(BuiltinDataType dataType) {
		assert moldManager != null;
		Validate.notNull(dataType);
		return moldManager.findDataTypeMold(dataType);
	}
	
	public BuiltinDataTypeMold findDataTypeMold(DataTypeCategory category) {
		assert moldManager != null;
		Validate.notNull(category);
		return moldManager.findDataTypeMold(category);
	}
	
	public BuiltinDataTypeMold findDataTypeMold(DataTypeCategory category, String typeName) {
		assert moldManager != null;
		Validate.notNull(category);
		Validate.notNull(typeName);
		return moldManager.findDataTypeMold(category, typeName);
	}
	
	public List<BuiltinDataTypeMold> getAllDataTypes() {
		assert moldManager != null;
		return moldManager.getTypeList();
	}
	
	public String getConnectionUriTemplate() {
		return "jdbc:";
	}
	
	public DataTypeResolver getDataTypeResolver() {
		assert dataTypeResolver != null;
		return dataTypeResolver;
	}
	
	public MoldManager getMoldManager() {
		return moldManager;
	}
	
	public Validator getValidator() {
		return new AllValidator();
	}
	
	public void importMetadata(RootModel rootModel, DatabaseMetaData meta, ImportMetadataConfig config,
			Map<String, EntityModel> importedEntities, Map<String, ForeignKey> importedForeignKeys) throws SQLException {
		Validate.notNull(rootModel);
		Validate.notNull(meta);
		Validate.notNull(config);
		Validate.notNull(importedEntities);
		Validate.notNull(importedForeignKeys);
		Connection connection = meta.getConnection();
		TypeSafeDatabaseMetaData metaData = new TypeSafeDatabaseMetaData(meta);
		DatabaseReader reader = new DatabaseReader(this, config, meta);
		
		advisor.setUpRead(connection);
		
		entityImportVisitor.intialize(metaData, rootModel, config, importedEntities, getDataTypeResolver());
		reader.readEnities(entityImportVisitor);
		
		foreignKeyImportVisitor.initialize(rootModel, importedForeignKeys);
		reader.readRelations(foreignKeyImportVisitor);
		
		advisor.tearDownRead(connection);
	}
	
	public RootModel parseStatements(List<SqlStatement> statements) {
		Validate.notNull(statements);
		throw new UnsupportedOperationException();
	}
	
	@Override
	public String toString() {
		return this.getClass().getName();
	}
}
