/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/03/02
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.facade;

import java.util.List;

import org.jiemamy.JiemamyProperty;
import org.jiemamy.ReferenceResolveException;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.KeyConstraint;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.dataset.RecordModel;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.index.IndexColumnModel;
import org.jiemamy.model.index.IndexModel;

/**
 * Jiemamyのモデル操作を単純化するためのファサードインターフェイス。
 * 
 * <p>ファサードの具体的な目的としては、以下の通り。
 * <ul>
 *   <li>不整合モデルが出来ないよう、適切な操作インターフェイスを提供する。</li>
 *   <li>モデルの編集に対するイベント発生をサポートする。</li>
 *   <li>{@link SavePoint}によるロールバックを提供する。</li>
 * </ul></p>
 * 
 * @since 0.2
 * @author daisuke
 */
public interface JiemamyFacade {
	
	/**
	 * {@link TableModel}に{@link AttributeModel}を追加する。
	 * 
	 * <ul>
	 *   <li>属性が既にこのテーブルに属している場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param tableModel 追加対象のテーブル
	 * @param attributeModel 追加する属性
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws IllegalArgumentException {@link KeyConstraint}追加時、対象テーブルに属さないカラムがキーカラムに含まれていた場合
	 * @throws ReferenceResolveException {@link KeyConstraint}追加時、キーカラムの参照が解決できなかった場合
	 * @since 0.2
	 */
	void addAttribute(TableModel tableModel, AttributeModel attributeModel);
	
	/**
	 * {@link TableModel}に{@link AttributeModel}を追加する。
	 * 
	 * <ul>
	 *   <li>属性が既にこのテーブルに属している場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param tableModel 追加対象のテーブル
	 * @param index 追加位置インデックス
	 * @param attributeModel 追加する属性
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws IllegalArgumentException {@link KeyConstraint}追加時、対象テーブルに属さないカラムがキーカラムに含まれていた場合
	 * @throws ReferenceResolveException {@link KeyConstraint}追加時、キーカラムの参照が解決できなかった場合
	 * @since 0.2
	 */
	void addAttribute(TableModel tableModel, int index, AttributeModel attributeModel);
	
	/**
	 * {@link RootModel}に{@link DataSetModel}を追加する。
	 * 
	 * <ul>
	 *   <li>データセットが既に追加済みの場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param dataSetModel 追加するデータセット
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addDataSet(DataSetModel dataSetModel);
	
	/**
	 * {@link RootModel}に{@link DataSetModel}を追加する。
	 * 
	 * <ul>
	 *   <li>データセットが既に追加済みの場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param dataSetModel 追加するデータセット
	 * @param index 挿入位置インデックス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addDataSet(DataSetModel dataSetModel, int index);
	
	/**
	 * {@link RootModel}に{@link DomainModel}を追加する。
	 * 
	 * <ul>
	 *   <li>ドメインが既に追加済みの場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param domainModel 追加するドメイン
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addDomain(DomainModel domainModel);
	
	/**
	 * {@link RootModel}に{@link DomainModel}を追加する。
	 * 
	 * <ul>
	 *   <li>ドメインが既に追加済みの場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param domainModel 追加するドメイン
	 * @param index 挿入位置インデックス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addDomain(DomainModel domainModel, int index);
	
	/**
	 * {@link RootModel}に{@link EntityModel}を追加する。
	 * 
	 * <ul>
	 *   <li>エンティティが既に追加済みの場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param entityModel 追加するエンティティ
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addEntity(EntityModel entityModel);
	
	/**
	 * {@link TableModel}に{@link IndexModel}を追加する。
	 * 
	 * <ul>
	 *   <li>インデックスが既にこのテーブルに属している場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param tableModel 追加対象のテーブル
	 * @param indexModel 追加するインデックス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addIndex(TableModel tableModel, IndexModel indexModel);
	
	/**
	 * {@link TableModel}に{@link IndexModel}を追加する。
	 * 
	 * <ul>
	 *   <li>インデックスが既にこのテーブルに属している場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param tableModel 追加対象のテーブル
	 * @param index 追加位置インデックス
	 * @param indexModel 追加するインデックス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addIndex(TableModel tableModel, int index, IndexModel indexModel);
	
	/**
	 * {@link IndexModel}に{@link IndexColumnModel}を追加する。
	 * 
	 * <ul>
	 *   <li>インデックスカラムが既にこのインデックスに属している場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param indexModel 追加対象のインデックス
	 * @param indexColumnModel 追加するインデックスカラム
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addIndexColumn(IndexModel indexModel, IndexColumnModel indexColumnModel);
	
	/**
	 * {@link IndexModel}に{@link IndexColumnModel}を追加する。
	 * 
	 * <ul>
	 *   <li>インデックスカラムが既にこのインデックスに属している場合は、何もしない</li>
	 * </ul>
	 * 
	 * @param indexModel 追加対象のインデックス
	 * @param index 挿入位置インデックス
	 * @param indexColumnModel 追加するインデックスカラム
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addIndexColumn(IndexModel indexModel, int index, IndexColumnModel indexColumnModel);
	
	/**
	 * {@link KeyConstraint}のキーカラム要素にカラムを追加する。
	 * 
	 * <ul>
	 *   <li>既に指定したカラムがキーカラムの要素であった場合は何もしない。</li>
	 * </ul>
	 * 
	 * @param keyConstraint 対象のキー制約
	 * @param columnModel 追加するカラム
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addKeyColumn(KeyConstraint keyConstraint, ColumnModel columnModel);
	
	/**
	 * {@link DataSetModel}に、指定したテーブル用のレコード保存領域を確保する。
	 * 
	 * <ul>
	 *   <li>既に指定したテーブル用のレコード保存領域が確保済みであった場合は何もしない。</li>
	 * </ul>
	 * 
	 * @param dataSetModel レコード追加対象の{@link DataSetModel}
	 * @param tableModel レコード追加対象の{@link TableModel}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addRecords(DataSetModel dataSetModel, TableModel tableModel);
	
	/**
	 * モデルが持つプロパティの値を変更する。
	 * 
	 * @param <T> 操作対象のモデルの型
	 * @param element 操作対象モデル
	 * @param property 変更するプロパティ
	 * @param newValue 新しい値 may be null
	 * @throws IllegalArgumentException 引数element, propertyに{@code null}を与えた場合
	 * @throws IllegalArgumentException propertyの型とnewValueの型が一致しなかった場合
	 * @since 0.2
	 */
	<T extends JiemamyElement>void changeModelProperty(T element, JiemamyProperty<T> property, Object newValue);
	
	/**
	 * {@link TableModel}から{@link AttributeModel}を削除する。
	 * 
	 * <ul>
	 *   <li>ローカルキーを削除する場合、そのキーを参照する外部キーも同時に削除する。</li>
	 *   <li>カラムを削除する場合、全てのキーに含まれる該当カラムへの参照も同時に削除する。</li>
	 * </ul>
	 * 
	 * @param tableModel 削除元テーブル
	 * @param attributeModel 削除対象の属性
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeAttribute(TableModel tableModel, AttributeModel attributeModel);
	
	/**
	 * {@link RootModel}から{@link DataSetModel}を削除する。
	 * 
	 * <ul>
	 *   <li>指定したデータセットが{@link RootModel}に属していなかった場合は何もしない。</li>
	 * </ul>
	 * 
	 * @param dataSetModel 削除するデータセット
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeDataSet(DataSetModel dataSetModel);
	
	/**
	 * {@link RootModel}から{@link DomainModel}を削除する。
	 * 
	 * <ul>
	 *   <li>指定したドメインが{@link RootModel}に属していなかった場合は何もしない。</li>
	 *   <li>該当ドメインへの参照は、ドメインに設定されていた型情報に書き換える。</li>
	 * </ul>
	 * 
	 * @param domainModel 削除するドメイン
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeDomain(DomainModel domainModel);
	
	/**
	 * {@link RootModel}から{@link EntityModel}を削除する。
	 * 
	 * <ul>
	 *   <li>指定したエンティティが{@link RootModel}に属していなかった場合は何もしない。</li>
	 *   <li>該当エンティティを参照する外部キーも同時に削除する。</li>
	 *   <li>{@link DataSetModel}が持つテーブルへの参照も同時に削除する。</li>
	 * </ul>
	 * 
	 * @param entityModel 削除するエンティティ
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeEntity(EntityModel entityModel);
	
	/**
	 * {@link TableModel}から{@link IndexModel}を削除する。
	 * 
	 * <ul>
	 *   <li>指定したインデックスがテーブルに属していなかった場合は何もしない。</li>
	 * </ul>
	 * 
	 * @param tableModel 削除元テーブル
	 * @param indexModel 削除するインデックス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeIndex(TableModel tableModel, IndexModel indexModel);
	
	/**
	 * {@link IndexModel}から{@link IndexColumnModel}を削除する。
	 * 
	 * <ul>
	 *   <li>指定したインデックスカラムがインデックスに属していなかった場合は何もしない。</li>
	 * </ul>
	 * 
	 * @param indexModel 削除元インデックス
	 * @param indexColumnModel 削除するインデックスカラム
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeIndexColumn(IndexModel indexModel, IndexColumnModel indexColumnModel);
	
	/**
	 * {@link KeyConstraint}のキーカラム要素からカラムを削除する。
	 * 
	 * <ul>
	 *   <li>キーカラム要素が、指定したカラムへの参照を保持していなかった場合は、何もしない。</li>
	 *   <li>削除の結果、キーカラムの数が0になった場合、キー制約自体を所属テーブルから削除する。</li>
	 *   <li>外部キーであった場合は、対応する参照カラムも削除する。</li>
	 * </ul>
	 * 
	 * @param keyConstraint 対象のキー制約
	 * @param columnModel 削除するカラム
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeKeyColumn(KeyConstraint keyConstraint, ColumnModel columnModel);
	
	/**
	 * {@link DataSetModel}に確保された、指定したテーブル用のレコード保存領域を削除する。
	 * 
	 * <ul>
	 *   <li>指定したテーブル用のレコード保存領域が存在しなかった場合は何もしない。</li>
	 * </ul>
	 * 
	 * @param dataSetModel レコード削除元の{@link DataSetModel}
	 * @param tableModel レコード削除対象の{@link TableModel}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeRecords(DataSetModel dataSetModel, TableModel tableModel);
	
	/**
	 * モデルの状態をセーブポイントまでロールバックする。
	 * 
	 * @param savePoint セーブポイント
	 * @throws IllegalArgumentException このファサードが発行したｓ{@link SavePoint}でない場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void rollback(SavePoint savePoint);
	
	/**
	 * 現在のモデルの状態にロールバックするためのセーブポイントを取得する。
	 * 
	 * @return セーブポイント
	 * @since 0.2
	 */
	SavePoint save();
	
	/**
	 * {@link RecordModel}の指定カラムのデータを設定する。
	 * 
	 * @param recordModel 対象レコード
	 * @param columnModel 対象カラム
	 * @param value データ
	 * @throws IllegalArgumentException 引数value<b>以外</b>に{@code null}を与えた場合
	 * @since 0.2
	 */
	void setRecordValue(RecordModel recordModel, ColumnModel columnModel, String value);
	
	/**
	 * リスト内の要素入れ替えを行う。
	 * 
	 * @param targetElement 入れ替え対象のリストを保持するモデル
	 * @param list 入れ替え対象のリスト
	 * @param subjectIndex 入れ替え要素１
	 * @param objectIndex 入れ替え要素２
	 * @throws IndexOutOfBoundsException リストに、指定したインデックスの要素が存在しない場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void swapListElement(JiemamyElement targetElement, List<?> list, int subjectIndex, int objectIndex);
	
	/**
	 * 外部キーのカラムマッピングを更新する。
	 * 
	 * @param foreignKey 対象の外部キー制約
	 * @param keyColumns 新しいキーカラムのリスト
	 * @param referenceColumns 新しい参照カラムのリスト
	 * @throws IllegalArgumentException 引数keyColumnsとreferenceColumnのサイズが異なる場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void updateForeignKeyMapping(ForeignKey foreignKey, List<ColumnRef> keyColumns, List<ColumnRef> referenceColumns);
}
