/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/02/22
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal;

import java.util.List;
import java.util.Map;

import org.apache.commons.lang.Validate;

import org.jiemamy.Jiemamy;
import org.jiemamy.ReferenceResolver;
import org.jiemamy.JiemamyViewProperty.NodeProfileProperty;
import org.jiemamy.editcommand.Command;
import org.jiemamy.exception.ElementNotFoundException;
import org.jiemamy.facade.JiemamyViewFacade;
import org.jiemamy.internal.editcommand.AddBendpointCommand;
import org.jiemamy.internal.editcommand.AddColumnToColumnRefListCommand;
import org.jiemamy.internal.editcommand.AddConnectionViewCommand;
import org.jiemamy.internal.editcommand.AddNodeViewCommand;
import org.jiemamy.internal.editcommand.ModifyModelPropertyCommand;
import org.jiemamy.internal.editcommand.MoveBendpointCommand;
import org.jiemamy.internal.editcommand.RemoveBendpointCommand;
import org.jiemamy.internal.editcommand.RemoveConnectionViewCommand;
import org.jiemamy.internal.editcommand.RemoveNodeViewCommand;
import org.jiemamy.internal.editcommand.ResetBendpointCommand;
import org.jiemamy.model.ConnectionProfile;
import org.jiemamy.model.DiagramPresentationModel;
import org.jiemamy.model.DiagramPresentations;
import org.jiemamy.model.NodeProfile;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.LocalKeyConstraint;
import org.jiemamy.model.connection.ConnectionAdapter;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.geometory.JmColor;
import org.jiemamy.model.geometory.JmPoint;
import org.jiemamy.model.geometory.JmRectangle;
import org.jiemamy.model.node.NodeAdapter;
import org.jiemamy.utils.CollectionsUtil;

/**
 * j-view用のEDITコマンド実行ファサードクラス。
 * 
 * @author daisuke
 * @author shin1ogawa
 */
public class JiemamyViewFacadeImpl extends JiemamyFacadeImpl implements JiemamyViewFacade {
	
	/**
	 * インスタンスを生成する。
	 * 
	 * @param jiemamy ファサードによる操作の対象となるコンテキスト
	 */
	public JiemamyViewFacadeImpl(Jiemamy jiemamy) {
		super(jiemamy);
	}
	
	public void addBendpoint(int diagramIndex, ConnectionAdapter connectionAdapter, int bendpointIndex, JmPoint position) {
		Validate.notNull(connectionAdapter);
		Validate.notNull(position);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		DiagramPresentations diagramPresentations = rootModel.getAdapter(DiagramPresentations.class);
		DiagramPresentationModel diagramPresentationModel = diagramPresentations.get(diagramIndex);
		ConnectionProfile connectionProfile = diagramPresentationModel.getConnectionProfiles().get(connectionAdapter);
		
		Command command =
				new AddBendpointCommand(eventBroker, diagramIndex, connectionProfile, bendpointIndex, position);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addConnectionProfile(int diagramIndex, ConnectionAdapter connectionAdapter,
			ConnectionProfile connectionProfile) {
		Validate.notNull(connectionAdapter);
		Validate.notNull(connectionProfile);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		Command command =
				new AddConnectionViewCommand(eventBroker, rootModel, diagramIndex, connectionAdapter, connectionProfile);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void addNode(int diagramIndex, NodeAdapter nodeAdapter, JmRectangle rectangle) {
		Validate.notNull(nodeAdapter);
		Validate.notNull(rectangle);
		if (nodeAdapter.unwrap() != null) {
			EntityModel entityModel = nodeAdapter.unwrap();
			super.addEntity(entityModel);
		}
		NodeProfile nodeProfile = jiemamy.getFactory().newModel(NodeProfile.class);
		addNodeProfile(diagramIndex, nodeAdapter, nodeProfile);
		changeNodeBoundary(diagramIndex, nodeAdapter, rectangle);
	}
	
	public void addNodeProfile(int diagramIndex, NodeAdapter nodeAdapter, NodeProfile nodeProfile) {
		Validate.notNull(nodeAdapter);
		Validate.notNull(nodeProfile);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		Command command = new AddNodeViewCommand(eventBroker, rootModel, diagramIndex, nodeAdapter, nodeProfile);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void changeNodeBoundary(int diagramIndex, NodeAdapter nodeAdapter, JmRectangle boundary) {
		Validate.notNull(nodeAdapter);
		Validate.notNull(boundary);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		DiagramPresentations diagramPresentations = rootModel.getAdapter(DiagramPresentations.class);
		DiagramPresentationModel diagramPresentationModel = diagramPresentations.get(diagramIndex);
		Map<NodeAdapter, NodeProfile> figureProfiles = diagramPresentationModel.getNodeProfiles();
		if (figureProfiles.containsKey(nodeAdapter) == false) {
			NodeProfile profile = jiemamy.getFactory().newModel(NodeProfile.class);
			figureProfiles.put(nodeAdapter, profile);
		}
		NodeProfile nodeProfile = figureProfiles.get(nodeAdapter);
		Command command =
				new ModifyModelPropertyCommand<NodeProfile>(eventBroker, nodeProfile, NodeProfileProperty.boundary,
						boundary);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void changeNodeColor(int diagramIndex, NodeAdapter nodeAdapter, JmColor color) {
		Validate.notNull(nodeAdapter);
		Validate.notNull(color);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		DiagramPresentations diagramPresentations = rootModel.getAdapter(DiagramPresentations.class);
		DiagramPresentationModel diagramPresentationModel = diagramPresentations.get(diagramIndex);
		Map<NodeAdapter, NodeProfile> figureProfiles = diagramPresentationModel.getNodeProfiles();
		if (figureProfiles.containsKey(nodeAdapter) == false) {
			NodeProfile profile = jiemamy.getFactory().newModel(NodeProfile.class);
			figureProfiles.put(nodeAdapter, profile);
		}
		NodeProfile nodeProfile = figureProfiles.get(nodeAdapter);
		Command command =
				new ModifyModelPropertyCommand<NodeProfile>(eventBroker, nodeProfile, NodeProfileProperty.color, color);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void createConnection(int diagramIndex, ConnectionAdapter connectionAdapter, TableModel sourceTable,
			TableModel targetTable) {
		Validate.notNull(connectionAdapter);
		Validate.notNull(sourceTable);
		Validate.notNull(targetTable);
		if (connectionAdapter.unwrap() != null) {
			ForeignKey foreignKey = connectionAdapter.unwrap();
			Jiemamy jiemamy = foreignKey.getJiemamy();
			ReferenceResolver referenceResolver = jiemamy.getReferenceResolver();
			
			LocalKeyConstraint key = null;
			try {
				key = targetTable.findPrimaryKey();
			} catch (ElementNotFoundException e) {
				List<LocalKeyConstraint> attributes = targetTable.findAttributes(LocalKeyConstraint.class);
				if (attributes.size() > 0) {
					key = attributes.get(0);
				}
			}
			
			List<ColumnModel> sourceColumns = sourceTable.findColumns();
			
			for (ColumnRef columnRef : key.getKeyColumns()) {
				ColumnModel keyColumn = sourceColumns.get(0);
				ColumnModel refColumn = referenceResolver.resolve(columnRef);
				
				Command command =
						new AddColumnToColumnRefListCommand(eventBroker, foreignKey, foreignKey.getKeyColumns(),
								keyColumn);
				command.execute();
				undoStack.push(command.getNegateCommand());
				command =
						new AddColumnToColumnRefListCommand(eventBroker, foreignKey, foreignKey.getReferenceColumns(),
								refColumn);
				command.execute();
				undoStack.push(command.getNegateCommand());
			}
			
			super.addAttribute(sourceTable, foreignKey);
		} else {
			// TODO 非wrapping connectionの処理
		}
		addConnectionProfile(diagramIndex, connectionAdapter, jiemamy.getFactory().newModel(ConnectionProfile.class));
	}
	
	public void moveBendpoint(int diagramIndex, ConnectionAdapter connectionAdapter, int bendpointIndex,
			JmPoint position) {
		Validate.notNull(connectionAdapter);
		Validate.notNull(position);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		DiagramPresentations diagramPresentations = rootModel.getAdapter(DiagramPresentations.class);
		DiagramPresentationModel diagramPresentationModel = diagramPresentations.get(diagramIndex);
		ConnectionProfile connectionProfile = diagramPresentationModel.getConnectionProfiles().get(connectionAdapter);
		
		Command command =
				new MoveBendpointCommand(eventBroker, diagramIndex, connectionProfile, bendpointIndex, position);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeBendpoint(int diagramIndex, ConnectionAdapter connectionAdapter, int bendpointIndex) {
		Validate.notNull(connectionAdapter);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		DiagramPresentations diagramPresentations = rootModel.getAdapter(DiagramPresentations.class);
		DiagramPresentationModel diagramPresentationModel = diagramPresentations.get(diagramIndex);
		ConnectionProfile connectionProfile = diagramPresentationModel.getConnectionProfiles().get(connectionAdapter);
		
		Command command = new RemoveBendpointCommand(eventBroker, diagramIndex, connectionProfile, bendpointIndex);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeConnection(int diagramIndex, ConnectionAdapter connectionAdapter) {
		Validate.notNull(connectionAdapter);
		
		ForeignKey foreignKey = connectionAdapter.unwrap();
		removeConnectionProfile(diagramIndex, connectionAdapter);
		EntityModel sourceTable = connectionAdapter.getSource().unwrap();
		if (sourceTable instanceof TableModel) {
			super.removeAttribute((TableModel) sourceTable, foreignKey);
		}
	}
	
	public void removeConnectionProfile(int diagramIndex, ConnectionAdapter connectionAdapter) {
		Validate.notNull(connectionAdapter);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		Command command = new RemoveConnectionViewCommand(eventBroker, rootModel, diagramIndex, connectionAdapter);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void removeNode(int diagramIndex, NodeAdapter nodeAdapter) {
		Validate.notNull(nodeAdapter);
		// 削除対象のモデルをソースとするコネクションのリスト、のコピー
		List<ConnectionAdapter> sourceConnections = CollectionsUtil.newArrayList(nodeAdapter.getSourceConnections());
		// 削除対象のモデルをターゲットとするコネクションのリスト、のコピー
		List<ConnectionAdapter> targetConnections = CollectionsUtil.newArrayList(nodeAdapter.getTargetConnections());
		
		for (ConnectionAdapter connection : sourceConnections) {
			removeConnection(diagramIndex, connection); // profileの削除だけではダメ。
		}
		
		for (ConnectionAdapter connection : targetConnections) {
			// DOCME ここはなぜ、profileの削除だけで、実体の削除を行わないのか？
			removeConnectionProfile(diagramIndex, connection);
		}
		
		removeNodeProfile(diagramIndex, nodeAdapter);
		if (nodeAdapter.unwrap() != null) {
			removeEntity(nodeAdapter.unwrap());
		}
	}
	
	public void removeNodeProfile(int diagramIndex, NodeAdapter nodeAdapter) {
		Validate.notNull(nodeAdapter);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		Command command = new RemoveNodeViewCommand(eventBroker, rootModel, diagramIndex, nodeAdapter);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void resetBendpoint(int diagramIndex, ConnectionAdapter connectionAdapter) {
		Validate.notNull(connectionAdapter);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		Command command = new ResetBendpointCommand(eventBroker, rootModel, diagramIndex, connectionAdapter);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
	
	public void setColor(int diagramIndex, NodeAdapter nodeAdapter, JmColor color) {
		Validate.notNull(nodeAdapter);
		Validate.notNull(color);
		RootModel rootModel = jiemamy.getFactory().getRootModel();
		DiagramPresentations diagramPresentations = rootModel.getAdapter(DiagramPresentations.class);
		DiagramPresentationModel diagramPresentationModel = diagramPresentations.get(diagramIndex);
		NodeProfile nodeProfile = diagramPresentationModel.getNodeProfiles().get(nodeAdapter);
		
		Command command =
				new ModifyModelPropertyCommand<NodeProfile>(eventBroker, nodeProfile, NodeProfileProperty.color, color);
		command.execute();
		undoStack.push(command.getNegateCommand());
	}
}
