/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/12/24
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.builder;

import java.util.List;

import org.apache.commons.lang.Validate;

import org.jiemamy.JiemamyFactory;
import org.jiemamy.ReferenceResolver;
import org.jiemamy.exception.ModelConsistenceException;
import org.jiemamy.experimental.builder.ForeignKeyBuilder;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.LocalKeyConstraint;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.utils.model.ForeignKeyUtil;

/**
 * {@link ForeignKey}を構築するためのビルダー実装クラス。
 * 
 * @author daisuke
 */
public class ForeignKeyBuilderImpl extends AbstractJiemamyBuilder<ForeignKey> implements ForeignKeyBuilder {
	
	private final TableModel constrainedTable;
	
	private final TableModel referenceTable;
	
	private List<LocalKeyConstraint> localKeys;
	
	private boolean complete;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param factory モデル生成のためのファクトリ
	 * @param constrainedTable 参照元テーブル
	 * @param referenceTable 参照先テーブル
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public ForeignKeyBuilderImpl(JiemamyFactory factory, TableModel constrainedTable, TableModel referenceTable) {
		super(factory, ForeignKey.class);
		Validate.notNull(constrainedTable);
		Validate.notNull(referenceTable);
		
		this.constrainedTable = constrainedTable;
		this.referenceTable = referenceTable;
		
		localKeys = referenceTable.findAttributes(LocalKeyConstraint.class);
	}
	
	public ForeignKeyBuilder mapping(ColumnModel keyColumn, ColumnModel referenceColumn) {
		return mapping(factory.newReference(keyColumn), factory.newReference(referenceColumn));
	}
	
	public ForeignKeyBuilder mapping(ColumnRef keyColumn, ColumnRef referenceColumn) {
		Validate.notNull(referenceColumn);
		Validate.notNull(keyColumn);
		
		ReferenceResolver resolver = constrainedTable.getJiemamy().getReferenceResolver();
		
		if (referenceTable.getAttributes().contains(resolver.resolve(referenceColumn)) == false) {
			throw new ModelConsistenceException();
		}
		if (constrainedTable.getAttributes().contains(resolver.resolve(keyColumn)) == false) {
			throw new ModelConsistenceException();
		}
		
		EntityModel referenceEntity = model.findReferencedEntity();
		if (referenceEntity != null) {
			ColumnModel columnModel = resolver.resolve(referenceColumn);
			TableModel definedTable = columnModel.findDeclaringTable();
			if (referenceEntity != definedTable) {
				throw new ModelConsistenceException();
			}
		}
		
		boolean valid = false;
		for (LocalKeyConstraint localKey : localKeys) {
			if (localKey.getKeyColumns().contains(referenceColumn)) {
				if (localKey.getKeyColumns().containsAll(model.getReferenceColumns())) {
					valid = true;
				}
			}
		}
		
		if (valid == false) {
			throw new ModelConsistenceException();
		}
		
		ForeignKeyUtil.addMapping(model, keyColumn, referenceColumn);
		
		return this;
	}
	
	@Override
	boolean isComplete() {
		// TODO Auto-generated method stub
		return complete;
	}
}
