/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/19
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.editcommand;

import java.lang.reflect.InvocationTargetException;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.Validate;

import org.jiemamy.EventBroker;
import org.jiemamy.JiemamyProperty;
import org.jiemamy.editcommand.Command;
import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.utils.model.JiemamyPropertyUtil;

/**
 * モデルの属性を変更するEDITコマンド。
 * 
 * @param <T> 属性を変更する対象のモデルの型
 * @author shin1ogawa
 * @author daisuke
 */
public class ModifyModelPropertyCommand<T extends JiemamyElement> extends AbstractCommand {
	
	/** 属性を変更する対象のモデル */
	private final T target;
	
	/** 変更する属性のフィールド */
	private final JiemamyProperty<T> property;
	
	/** 変更前の値 */
	private final Object oldValue;
	
	/** 変更後の値 */
	private final Object newValue;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param eventBroker イベント通知用{@link EventBroker}
	 * @param target 属性を変更する対象のモデル
	 * @param property 変更する属性のフィールド
	 * @param newValue 変更後の値 may be null
	 * @throws IllegalArgumentException 引数target, propertyに{@code null}を与えた場合
	 */
	public ModifyModelPropertyCommand(EventBroker eventBroker, T target, JiemamyProperty<T> property, Object newValue) {
		this(eventBroker, target, property, newValue, null);
	}
	
	/**
	 * インスタンスを生成する。
	 * 
	 * @param eventBroker イベント通知用{@link EventBroker}
	 * @param target 属性を変更する対象のモデル
	 * @param property 変更する属性のフィールド
	 * @param newValue 変更後の値 may be null
	 * @param oldValue 変更前の値 may be null
	 * @throws IllegalArgumentException 引数target, propertyに{@code null}を与えた場合
	 */
	public ModifyModelPropertyCommand(EventBroker eventBroker, T target, JiemamyProperty<T> property, Object newValue,
			Object oldValue) {
		super(eventBroker);
		Validate.notNull(target);
		Validate.notNull(property);
		this.target = target;
		this.property = property;
		this.newValue = newValue;
		if (oldValue != null) {
			this.oldValue = oldValue;
		} else {
			this.oldValue = JiemamyPropertyUtil.getProperty(target, property);
		}
	}
	
	public Command getNegateCommand() {
		return new ModifyModelPropertyCommand<T>(getEventBroker(), target, property, oldValue, newValue);
	}
	
	/**
	 * 属性を変更する対象のモデルを取得する。
	 * 
	 * @return 属性を変更する対象のモデル
	 */
	public JiemamyElement getTarget() {
		return target;
	}
	
	@Override
	public String toString() {
		return "ModifyModelPropertyCommand(" + target + "." + property + "=" + oldValue + " to " + newValue + ")";
	}
	
	@Override
	protected void execute0() {
		String propertyName = property.name();
		try {
			PropertyUtils.setProperty(target, propertyName, newValue);
		} catch (IllegalAccessException e) {
			throw new UnexpectedConditionError("JiemamyPropertyで表されるプロパティにアクセスできない", e);
		} catch (InvocationTargetException e) {
			throw new UnexpectedConditionError(propertyName + "に対するgetter/setterが存在しない?", e);
		} catch (NoSuchMethodException e) {
			throw new UnexpectedConditionError(propertyName + "に対するgetter/setterが存在しない?", e);
		}
	}
}
