/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/21
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.editcommand;

import java.util.List;

import org.apache.commons.lang.Validate;

import org.jiemamy.EventBroker;
import org.jiemamy.editcommand.Command;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;

/**
 * {@link ColumnRef}のリストから{@link ColumnModel}を削除するEDITコマンド。
 * 
 * @author shin1ogawa
 */
public class RemoveColumnFromColumnRefListCommand extends AbstractCommand {
	
	/** カラム参照リストを保持するモデル */
	private final JiemamyElement target;
	
	/** 削除元のリスト */
	private final List<ColumnRef> columnRefList;
	
	/** 削除されるカラム */
	private final ColumnModel columnModel;
	
	/** 削除位置インデックス */
	private final int index;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param eventBroker イベント通知用{@link EventBroker}
	 * @param target カラム参照リストを保持するモデル
	 * @param columnRefList 削除対象のリスト
	 * @param columnModel 削除されるカラム
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public RemoveColumnFromColumnRefListCommand(EventBroker eventBroker, JiemamyElement target,
			List<ColumnRef> columnRefList, ColumnModel columnModel) {
		super(eventBroker);
		Validate.notNull(target);
		Validate.notNull(columnRefList);
		Validate.notNull(columnModel);
		
		// List内での削除対象カラムのindexを探す。
		int i = 0;
		for (ColumnRef ref : columnRefList) {
			if (ref.getReferenceId().equals(columnModel.getId())) {
				break;
			}
			i++;
		}
		if (i >= columnRefList.size()) {
			// List内に削除対象カラムが見つからない。
			throw new IllegalArgumentException();
		}
		index = i;
		this.target = target;
		this.columnRefList = columnRefList;
		this.columnModel = columnModel;
	}
	
	/**
	 * インスタンスを生成する。
	 * 
	 * @param eventBroker イベント通知用{@link EventBroker}
	 * @param target 操作対象の{@link JiemamyElement}
	 * @param columnRefList 削除対象のリスト
	 * @param columnModel 削除されるカラム
	 * @param index 追加位置をあらわすインデックス値
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public RemoveColumnFromColumnRefListCommand(EventBroker eventBroker, JiemamyElement target,
			List<ColumnRef> columnRefList, ColumnModel columnModel, int index) {
		super(eventBroker);
		Validate.notNull(target);
		Validate.notNull(columnRefList);
		Validate.notNull(columnModel);
		this.target = target;
		this.columnRefList = columnRefList;
		this.columnModel = columnModel;
		this.index = index;
	}
	
	public Command getNegateCommand() {
		return new AddColumnToColumnRefListCommand(getEventBroker(), target, columnRefList, columnModel, index);
	}
	
	public JiemamyElement getTarget() {
		return target;
	}
	
	@Override
	protected void execute0() {
		columnRefList.remove(index);
	}
}
