/*
 * Copyright 2009 Jiemamy Project and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.enhance;

import java.util.Map;
import java.util.UUID;

import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.jiemamy.Jiemamy;
import org.jiemamy.JiemamyFactory;
import org.jiemamy.internal.serializer.JiemamyDomSerializer;
import org.jiemamy.model.ConnectionProfile;
import org.jiemamy.model.DiagramPresentationModel;
import org.jiemamy.model.DiagramPresentations;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.Level;
import org.jiemamy.model.Mode;
import org.jiemamy.model.NodeProfile;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.connection.ConnectionAdapter;
import org.jiemamy.model.geometory.JmColor;
import org.jiemamy.model.geometory.JmPoint;
import org.jiemamy.model.geometory.JmRectangle;
import org.jiemamy.model.node.NodeAdapter;
import org.jiemamy.model.node.StickyModel;
import org.jiemamy.serializer.SerializationException;
import org.jiemamy.utils.DomUtil;
import org.jiemamy.utils.IterableNodeList;
import org.jiemamy.utils.LogMarker;
import org.jiemamy.utils.SmartXpath;
import org.jiemamy.utils.UUIDUtil;
import org.jiemamy.utils.enhancer.Invocation;
import org.jiemamy.utils.enhancer.InvocationHandler;
import org.jiemamy.utils.model.RootModelUtil;
import org.jiemamy.xml.CoreQName;
import org.jiemamy.xml.JiemamyQName;
import org.jiemamy.xml.ViewNamespace;
import org.jiemamy.xml.ViewQName;

/**
 * {@link JiemamyFactory#getSerializer()}を拡張する。
 * 
 * @author daisuke
 */
public class ViewDomSerializerEnhancer implements InvocationHandler {
	
	public Object handle(Invocation invocation) throws Throwable {
		// new FactoryExtension() -> new ExtensionEnhancer.ViewExtention();
		Jiemamy jiemamy = (Jiemamy) invocation.getArguments()[0];
		return new ViewDomSerializerEnhancer.JiemamyViewDomSerializer(jiemamy);
	}
	

	/**
	 * {@link JiemamyDomSerializer}の拡張。
	 * 
	 * @author daisuke
	 */
	public static class JiemamyViewDomSerializer extends JiemamyDomSerializer {
		
		private static Logger logger = LoggerFactory.getLogger(JiemamyViewDomSerializer.class);
		
		private Jiemamy jiemamy;
		
		private SmartXpath smartXpath;
		

		/**
		 * インスタンスを生成する。
		 * 
		 * @param jiemamy コンテキスト
		 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
		 */
		public JiemamyViewDomSerializer(Jiemamy jiemamy) {
			super(jiemamy);
			this.jiemamy = jiemamy;
			smartXpath = new SmartXpath(jiemamy);
			logger.info(LogMarker.LIFECYCLE, "JiemamyDomSerializer extended by Jiemamy VIEW Specification");
		}
		
		@Override
		protected Document toDom(RootModel rootModel) throws ParserConfigurationException {
			Document document = super.toDom(rootModel);
			
			DiagramPresentations diagramPresentations = rootModel.getAdapter(DiagramPresentations.class);
			if (diagramPresentations == null || diagramPresentations.size() == 0) {
				return document;
			}
			
			Element rootElement = document.getDocumentElement();
			StringBuilder sb = new StringBuilder();
			sb.append(rootElement.getAttribute("xsi:schemaLocation"));
			sb.append(" ");
			sb.append(ViewNamespace.NS_VIEW.getNamespaceURI().toString());
			sb.append(" ");
			sb.append(ViewNamespace.NS_VIEW.getXmlSchemaLocation());
			rootElement.setAttribute("xsi:schemaLocation", sb.toString());
			
			Element diagramPresentationsElement =
					DomUtil.newChild(document.getDocumentElement(), ViewQName.DIAGRAM_PRESENTATIONS);
			
			logger.trace("diagramPresentations");
			for (DiagramPresentationModel presentation : diagramPresentations) {
				logger.trace("  diagramPresentation: id=" + presentation.getId());
				Element presentationElement =
						DomUtil.newChild(diagramPresentationsElement, ViewQName.DIAGRAM_PRESENTATION);
				presentationElement.setAttribute("id", presentation.getId().toString());
				
				logger.trace("    name: " + presentation.getName());
				DomUtil.newChild(presentationElement, CoreQName.NAME, presentation.getName());
				logger.trace("    mode: " + presentation.getMode());
				DomUtil.newChild(presentationElement, ViewQName.MODE, presentation.getMode());
				logger.trace("    level: " + presentation.getLevel());
				DomUtil.newChild(presentationElement, ViewQName.LEVEL, presentation.getLevel());
				logger.trace("    nodeProfiles");
				Element nodeProfilesElement = DomUtil.newChild(presentationElement, ViewQName.NODE_PROFILES);
				for (Map.Entry<NodeAdapter, NodeProfile> entry : presentation.getNodeProfiles().entrySet()) {
					NodeAdapter node = entry.getKey();
					NodeProfile nodeProfile = entry.getValue();
					logger.trace("      nodeProfile");
					Element nodeProfileElement = DomUtil.newChild(nodeProfilesElement, ViewQName.NODE_PROFILE);
					nodeProfileElement.setAttribute("id", nodeProfile.getId().toString());
					Element nodeElement;
					if (node.unwrap() == null) {
						nodeElement = DomUtil.newChild(nodeProfileElement, ViewQName.STICKY);
						StickyModel stickyModel = (StickyModel) node;
						nodeElement.setAttribute("id", stickyModel.getId().toString());
						DomUtil.newChild(nodeElement, ViewQName.CONTENTS, stickyModel.getContents());
						logger.trace("        sticky: " + stickyModel.getContents());
					} else {
						nodeElement = DomUtil.newChild(nodeProfileElement, ViewQName.NODE_OBJECT_REF);
						nodeElement.setAttribute("ref", node.unwrap().getId().toString());
						logger.trace("        node ref: " + node.unwrap().getId().toString());
					}
					Element boundaryElement = DomUtil.newChild(nodeProfileElement, ViewQName.BOUNDARY);
					boundaryElement.setAttribute("x", Integer.toString(nodeProfile.getBoundary().x));
					boundaryElement.setAttribute("y", Integer.toString(nodeProfile.getBoundary().y));
					boundaryElement.setAttribute("width", Integer.toString(nodeProfile.getBoundary().width));
					boundaryElement.setAttribute("height", Integer.toString(nodeProfile.getBoundary().height));
					JmColor color = nodeProfile.getColor();
					if (color != null) {
						boundaryElement.setAttribute("color", color.toString());
					}
					logger.trace("        boundary: " + boundaryElement);
				}
				Element connectionProfilesElement =
						DomUtil.newChild(presentationElement, ViewQName.CONNECTION_PROFILES);
				logger.trace("    connectionProfiles");
				for (Map.Entry<ConnectionAdapter, ConnectionProfile> entry : presentation.getConnectionProfiles()
					.entrySet()) {
					logger.trace("      connectionProfile");
					ConnectionAdapter connection = entry.getKey();
					ConnectionProfile connectionProfile = entry.getValue();
					Element connectionProfileElement =
							DomUtil.newChild(connectionProfilesElement, ViewQName.CONNECTION_PROFILE);
					connectionProfileElement.setAttribute("id", connectionProfile.getId().toString());
					if (connection.unwrap() == null) {
						logger.warn("      Unknown connection: " + connection.getClass().getName());
						continue;
					} else {
						Element connectionObjectRef =
								DomUtil.newChild(connectionProfileElement, ViewQName.CONNECTION_OBJECT_REF);
						connectionObjectRef.setAttribute("ref", connection.unwrap().getId().toString());
						logger.trace("      FK");
					}
					if (connectionProfile.getBendpoints().size() > 0) {
						Element bendpointsElement = DomUtil.newChild(connectionProfileElement, ViewQName.BENDPOINTS);
						logger.trace("      bendpoints");
						for (JmPoint point : connectionProfile.getBendpoints()) {
							Element bendpointElement = DomUtil.newChild(bendpointsElement, ViewQName.BENDPOINT);
							bendpointElement.setAttribute("x", Integer.toString(point.x));
							bendpointElement.setAttribute("y", Integer.toString(point.y));
							logger.trace("        bendpoint: " + bendpointElement);
						}
					}
				}
			}
			
			return document;
		}
		
		@Override
		protected RootModel toRootModel(Document document) throws SerializationException {
			RootModel rootModel = super.toRootModel(document);
			NodeList dpNodeList =
					smartXpath.getNodes(document, "/core:jiemamy/view:diagramPresentations/view:diagramPresentation");
			Iterable<Node> dpNodeListIterable = new IterableNodeList(dpNodeList);
			logger.debug("presentation count = " + dpNodeList.getLength());
			
			JiemamyFactory factory = jiemamy.getFactory();
			
			for (Node dpNode : dpNodeListIterable) {
				if ((dpNode instanceof Element) == false) {
					continue;
				}
				Element dpElement = (Element) dpNode;
				DiagramPresentationModel presentation =
						factory.newModel(DiagramPresentationModel.class, getUUID(dpElement, CoreQName.ID));
				presentation.setName(smartXpath.getTextContent(dpElement, "core:name"));
				
				String modeString = smartXpath.getTextContent(dpElement, "view:mode");
				if (StringUtils.isEmpty(modeString) == false) {
					presentation.setMode(Mode.valueOf(modeString));
				}
				
				String levelString = smartXpath.getTextContent(dpElement, "view:level");
				if (StringUtils.isEmpty(levelString) == false) {
					presentation.setLevel(Level.valueOf(levelString));
				}
				
				processNodeProfile(factory, presentation, dpElement);
				processConnectionProfile(factory, presentation, dpElement);
				
				rootModel.getAdapter(DiagramPresentations.class).add(presentation);
			}
			return rootModel;
		}
		
		/**
		 * 要素の属性に設定された値を{@link UUID}として取得する。
		 * 
		 * @param domNode ノード
		 * @param node 属性指定
		 * @return {@link UUID}
		 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
		 */
		private UUID getUUID(Node domNode, JiemamyQName node) {
			Validate.notNull(node);
			
			String uuidStr = smartXpath.getAttr(domNode, "@" + node.getQNameString()).getValue();
			return UUIDUtil.valueOfOrRandom(uuidStr);
		}
		
		private void processConnectionProfile(JiemamyFactory factory, DiagramPresentationModel presentation,
				Element dpElement) {
			RootModel rootModel = factory.getRootModel();
			NodeList connectionProfileNodeList =
					smartXpath.getNodes(dpElement, "view:connectionProfiles/view:connectionProfile");
			Iterable<Node> connectionProfileNodeListIterable = new IterableNodeList(connectionProfileNodeList);
			for (Node connectionProfileNode : connectionProfileNodeListIterable) {
				ConnectionAdapter connectionAdapter;
				Node objectRefNode = smartXpath.getNode(connectionProfileNode, "view:connectionObjectRef");
				if (objectRefNode != null) {
					UUID refId = getUUID(objectRefNode, CoreQName.REF);
					JiemamyElement jElement = RootModelUtil.findJiemamyElement(rootModel, refId);
					if (jElement == null) {
						logger.error("null element: " + refId.toString());
						continue;
					}
					connectionAdapter = jElement.getAdapter(ConnectionAdapter.class);
					if (connectionAdapter == null) {
						logger.error("null connection: " + jElement.getId().toString());
						continue;
					}
				} else {
					continue;
				}
				ConnectionProfile connectionProfile =
						factory.newModel(ConnectionProfile.class, getUUID(connectionProfileNode, CoreQName.ID));
				
				NodeList bendpointNodeList =
						smartXpath.getNodes(connectionProfileNode, "view:bendpoints/view:bendpoint");
				Iterable<Node> bendpointNodeListIterable = new IterableNodeList(bendpointNodeList);
				for (Node bendpointNode : bendpointNodeListIterable) {
					int x = Integer.valueOf(smartXpath.getAttr(bendpointNode, "@x").getValue());
					int y = Integer.valueOf(smartXpath.getAttr(bendpointNode, "@y").getValue());
					connectionProfile.getBendpoints().add(new JmPoint(x, y));
				}
				presentation.getConnectionProfiles().put(connectionAdapter, connectionProfile);
			}
		}
		
		private void processNodeProfile(JiemamyFactory factory, DiagramPresentationModel presentation, Element dpElement) {
			RootModel rootModel = factory.getRootModel();
			NodeList nodeProfileNodeList = smartXpath.getNodes(dpElement, "view:nodeProfiles/view:nodeProfile");
			Iterable<Node> nodeProfileNodeListIterable = new IterableNodeList(nodeProfileNodeList);
			for (Node nodeProfileNode : nodeProfileNodeListIterable) {
				NodeAdapter nodeAdapter;
				Node objectRefNode = smartXpath.getNode(nodeProfileNode, "view:nodeObjectRef");
				if (objectRefNode != null) {
					UUID refId = getUUID(objectRefNode, CoreQName.REF);
					JiemamyElement jElement = RootModelUtil.findJiemamyElement(rootModel, refId);
					if (jElement == null) {
						logger.error("null element: " + refId.toString());
						continue;
					}
					nodeAdapter = jElement.getAdapter(NodeAdapter.class);
					if (nodeAdapter == null) {
						logger.error("null node: " + jElement.getId().toString());
						continue;
					}
				} else {
					Node stickytNode = smartXpath.getNode(nodeProfileNode, "view:sticky");
					StickyModel stickyModel = factory.newModel(StickyModel.class, getUUID(stickytNode, CoreQName.ID));
					stickyModel.setContents(smartXpath.getTextContent(stickytNode, "view:contents"));
					nodeAdapter = stickyModel;
				}
				int x = Integer.valueOf(smartXpath.getAttr(nodeProfileNode, "view:boundary/@x").getValue());
				int y = Integer.valueOf(smartXpath.getAttr(nodeProfileNode, "view:boundary/@y").getValue());
				int width = Integer.valueOf(smartXpath.getAttr(nodeProfileNode, "view:boundary/@width").getValue());
				int height = Integer.valueOf(smartXpath.getAttr(nodeProfileNode, "view:boundary/@height").getValue());
				
				Attr colorAttr = smartXpath.getAttr(nodeProfileNode, "view:boundary/@color");
				JmColor color = colorAttr == null ? null : JmColor.parse(colorAttr.getValue());
				
				NodeProfile nodeProfile = factory.newModel(NodeProfile.class, getUUID(nodeProfileNode, CoreQName.ID));
				nodeProfile.setBoundary(new JmRectangle(x, y, width, height));
				nodeProfile.setColor(color);
				presentation.getNodeProfiles().put(nodeAdapter, nodeProfile);
			}
			presentation.setName(smartXpath.getTextContent(dpElement, "core:name"));
		}
	}
}
