/*
 * Copyright 2009 Jiemamy Project and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.enhance;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.TreeMap;
import java.util.UUID;

import org.jiemamy.Jiemamy;
import org.jiemamy.JiemamyFactory;
import org.jiemamy.internal.model.ConnectionProfileImpl;
import org.jiemamy.internal.model.DiagramPresentationModelImpl;
import org.jiemamy.internal.model.DiagramPresentationsImpl;
import org.jiemamy.internal.model.NodeProfileImpl;
import org.jiemamy.internal.model.connection.ConnectionAdapterImpl;
import org.jiemamy.internal.model.node.NodeAdapterImpl;
import org.jiemamy.internal.model.node.StickyModelImpl;
import org.jiemamy.model.ConnectionProfile;
import org.jiemamy.model.DiagramPresentationModel;
import org.jiemamy.model.NodeProfile;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.connection.ConnectionAdapter;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.geometory.JmPoint;
import org.jiemamy.model.node.NodeAdapter;
import org.jiemamy.model.node.StickyModel;
import org.jiemamy.utils.enhancer.Invocation;
import org.jiemamy.utils.enhancer.InvocationHandler;

/**
 * {@link JiemamyFactory#newModel(Class)}を拡張する。
 * 
 * @author daisuke
 */
public class ViewFactoryExtensionEnhancer implements InvocationHandler {
	
	@SuppressWarnings("unchecked")
	public Object handle(Invocation invocation) throws Throwable {
		// new FactoryExtension() -> new ViewFactoryExtensionEnhancer.ViewExtention();
		return new ViewFactoryExtensionEnhancer.ViewExtention();
	}
	

	/**
	 * {@link FactoryExtension}の拡張。
	 * 
	 * @param <T> 生成したいインスタンスの型
	 * @author daisuke
	 */
	public static class ViewExtention<T> extends FactoryExtension<T> {
		
		@Override
		public T newInstance(Class<? extends T> clazz, Jiemamy jiemamy, UUID id) {
			T product = super.newInstance(clazz, jiemamy, id);
			if (RootModel.class.isAssignableFrom(clazz)) {
				DiagramPresentationsImpl diagramPresentations = new DiagramPresentationsImpl((RootModel) product);
				diagramPresentations.setPresentations(new ArrayList<DiagramPresentationModel>());
				((RootModel) product).registerAdapter(diagramPresentations);
			} else if (EntityModel.class.isAssignableFrom(clazz)) {
				assert product != null;
				NodeAdapterImpl node = new NodeAdapterImpl((EntityModel) product);
				((EntityModel) product).registerAdapter(node);
			} else if (ForeignKey.class.isAssignableFrom(clazz)) {
				assert product != null;
				ConnectionAdapterImpl connection = new ConnectionAdapterImpl((ForeignKey) product);
				((ForeignKey) product).registerAdapter(connection);
			} else if (clazz == DiagramPresentationModel.class) {
				assert product == null;
				DiagramPresentationModelImpl presentation = new DiagramPresentationModelImpl(jiemamy, id);
				presentation.setConnectionProfiles(new TreeMap<ConnectionAdapter, ConnectionProfile>(
						ConnectionComparator.INSTANCE));
				presentation.setNodeProfiles(new TreeMap<NodeAdapter, NodeProfile>(NodeComparator.INSTANCE));
				product = clazz.cast(presentation);
			} else if (clazz == NodeProfile.class) {
				assert product == null;
				NodeProfileImpl nodeProfile = new NodeProfileImpl(jiemamy, id);
				product = clazz.cast(nodeProfile);
			} else if (clazz == ConnectionProfile.class) {
				assert product == null;
				ConnectionProfileImpl connectionProfile = new ConnectionProfileImpl(jiemamy, id);
				connectionProfile.setBendpoints(new ArrayList<JmPoint>());
				product = clazz.cast(connectionProfile);
			} else if (clazz == StickyModel.class) {
				assert product == null;
				StickyModelImpl stickyModel = new StickyModelImpl(jiemamy, id);
				product = clazz.cast(stickyModel);
			}
			return product;
		}
	}
	
	/**
	 * コネクションのソートオーダーを決めるコンパレータ。
	 * 
	 * @author daisuke
	 */
	protected enum ConnectionComparator implements Comparator<ConnectionAdapter> {
		
		/** singleton instance */
		INSTANCE;
		
		public int compare(ConnectionAdapter o1, ConnectionAdapter o2) {
			if (o1 == o2) {
				return 0;
			} else if (o1 == null) {
				return -1;
			} else if (o2 == null) {
				return 1;
			}
			
			ForeignKey f1 = o1.unwrap();
			ForeignKey f2 = o2.unwrap();
			if (f1 != null && f2 != null) {
				return FactoryExtension.JiemamyElementComparator.INSTANCE.compare(f1, f2);
			} else if (f1 != null) {
				return 1;
			} else if (f2 != null) {
				return -1;
			} else {
				return 0;
			}
		}
	}
	
	/**
	 * エンティティのソートオーダーを決めるコンパレータ。
	 * 
	 * <p>まずFQCNで比較し、同一だったらUUIDで比較。</p>
	 * 
	 * @author daisuke
	 */
	protected enum NodeComparator implements Comparator<NodeAdapter> {
		
		/** singleton instance */
		INSTANCE;
		
		public int compare(NodeAdapter o1, NodeAdapter o2) {
			Integer result = null;
			if (o1 == o2) {
				result = 0;
			} else if (o1 == null) {
				result = -1;
			} else if (o2 == null) {
				result = 1;
			} else if (o1.unwrap() == null && o2.unwrap() == null) {
				result = o1.getClass().getName().compareTo(o2.getClass().getName());
			} else if (o1.unwrap() == null) {
				result = 1;
			} else if (o2.unwrap() == null) {
				result = -1;
			}
			
			if (result != null) {
				return result;
			}
			
			EntityModel e1 = o1.unwrap();
			EntityModel e2 = o2.unwrap();
			if (e1 != null && e2 != null) {
				return FactoryExtension.JiemamyElementComparator.INSTANCE.compare(e1, e2);
			} else if (e1 != null) {
				return 1;
			} else if (e2 != null) {
				return -1;
			} else {
				return 0;
			}
		}
	}
	
}
