/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/06/09
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.model;

import java.util.Collection;
import java.util.List;
import java.util.SortedSet;
import java.util.UUID;

import org.apache.commons.lang.Validate;

import org.jiemamy.Jiemamy;
import org.jiemamy.ReferenceResolver;
import org.jiemamy.dialect.Dialect;
import org.jiemamy.exception.ElementNotFoundException;
import org.jiemamy.exception.TooManyElementsException;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.utils.ValidateUtil;

/**
 * 1つのDBスキーマ・データ全体を表す。
 * 
 * @author daisuke
 */
public class RootModelImpl extends AbstractJiemamyElement implements RootModel {
	
	/** SQL方言ID */
	private String dialectClassName;
	
	/** 開始スクリプト */
	private String beginScript;
	
	/** 終了スクリプト */
	private String endScript;
	
	/** 説明文 */
	private String description;
	
	/** スキーマ名 */
	private String schemaName;
	
	/** ドメイン定義のリスト */
	private List<DomainModel> domains;
	
	/** エンティティ定義のセット */
	private SortedSet<EntityModel> entities;
	
	/**
	 * INSERTデータセットのリスト
	 * 
	 * bacchusでは、1つのDBモデルに対して、INSERTするデータセットが1つしか定義できなかったが、
	 * データセットモデルを分離してCollectionで持つことにより、複数のデータセットを保持できる。
	 * （テストA用のデータセット、テストB用のデータセット... 等）
	 */
	private List<DataSetModel> dataSets;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param jiemamy コンテキスト
	 * @param id モデルID
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public RootModelImpl(Jiemamy jiemamy, UUID id) {
		super(jiemamy, id);
	}
	
//	@Override
//	public RootModel clone() {
//		try {
//			return (RootModel) super.clone();
//		} catch (CloneNotSupportedException e) {
//			throw new UnexpectedConditionError("RootModel does not implement Cloneable interface.");
//		}
//	}
	
	public Dialect findDialect() throws ClassNotFoundException {
		return getJiemamy().getDialect(this);
	}
	
	public <T extends EntityModel>Collection<T> findEntities(Class<T> clazz) {
		Validate.notNull(clazz);
		
		List<T> nodes = CollectionsUtil.newArrayList();
		for (EntityModel entityModel : getEntities()) {
			if (clazz.isAssignableFrom(entityModel.getClass())) {
				// 直前でロジックによる型チェックを行っているため、キャスト安全である。
				@SuppressWarnings("unchecked")
				T castedNode = (T) entityModel;
				nodes.add(castedNode);
			}
		}
		return nodes;
	}
	
	public String getBeginScript() {
		return beginScript;
	}
	
	public List<DataSetModel> getDataSets() {
		assert dataSets != null;
		return dataSets;
	}
	
	public String getDescription() {
		return description;
	}
	
	public String getDialectClassName() {
		return dialectClassName;
	}
	
	public List<DomainModel> getDomains() {
		assert domains != null;
		return domains;
	}
	
	public String getEndScript() {
		return endScript;
	}
	
	public SortedSet<EntityModel> getEntities() {
		assert entities != null;
		return entities;
	}
	
	public <T extends EntityModel>T getEntity(Class<T> clazz, String entityName) {
		Validate.notNull(entityName);
		
		List<T> entities = CollectionsUtil.newArrayList();
		for (EntityModel entityModel : getEntities()) {
			if (entityName.equalsIgnoreCase(entityModel.getName())
					&& (clazz == null || clazz.isAssignableFrom(entityModel.getClass()))) {
				entities.add((T) entityModel);
			}
		}
		if (entities.size() == 0) {
			throw new ElementNotFoundException(this, entityName);
		} else if (entities.size() > 1) {
			throw new TooManyElementsException(this, entityName, entities);
		}
		return entities.get(0);
	}
	
	public String getSchemaName() {
		return schemaName;
	}
	
	public void normalize() {
		ReferenceResolver resolver = getJiemamy().getReferenceResolver();
		for (TableModel tableModel : findEntities(TableModel.class)) {
			// PKの正規化
			try {
				tableModel.findAttribute(PrimaryKey.class, false);
			} catch (ElementNotFoundException e) {
				try {
					PrimaryKey primaryKey = tableModel.findPrimaryKey();
					ColumnRef columnRef = primaryKey.getKeyColumns().get(0);
					resolver.resolve(columnRef).setPrimaryKey(null);
					tableModel.getAttributes().add(primaryKey);
				} catch (ElementNotFoundException e2) {
					// nothing to do
				}
			}
			
			// UKの正規化
			for (ColumnModel columnModel : tableModel.findColumns()) {
				UniqueKey uniqueKey = columnModel.getUniqueKey();
				if (uniqueKey != null) {
					columnModel.setUniqueKey(null);
					tableModel.getAttributes().add(uniqueKey);
				}
			}
			
			// 属性順序の正規化
			// THINK 順序の正規化をするかしないか
			//			Comparator<AttributeModel> comparator = new AttributeComparator(tableModel);
			//			Collections.sort(tableModel.getAttributes(), comparator);
		}
	}
	
	public void setBeginScript(String beginScript) {
		this.beginScript = beginScript;
	}
	
	/**
	 * INSERTデータセットを設定する。
	 * 
	 * @param insertDataSets INSERTデータセット
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public void setDataSets(List<DataSetModel> insertDataSets) {
		Validate.notNull(insertDataSets);
		ValidateUtil.injectionSetter(dataSets);
		dataSets = insertDataSets;
	}
	
	public void setDescription(String description) {
		this.description = description;
	}
	
	public void setDialectClassName(String dialectClassName) {
		this.dialectClassName = dialectClassName;
	}
	
	/**
	 * ドメインのリストを設定する。
	 * 
	 * @param domains ドメインのリスト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public void setDomains(List<DomainModel> domains) {
		Validate.notNull(domains);
		ValidateUtil.injectionSetter(this.domains);
		this.domains = domains;
	}
	
	public void setEndScript(String endScript) {
		this.endScript = endScript;
	}
	
	/**
	 * ノードのリストを設定する。
	 * 
	 * @param entities ノードのリスト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public void setEntities(SortedSet<EntityModel> entities) {
		Validate.notNull(entities);
		ValidateUtil.injectionSetter(this.entities);
		this.entities = entities;
	}
	
	public void setSchemaName(String schemaName) {
		this.schemaName = schemaName;
	}
	
	@Override
	public String toString() {
		return getClass().getName() + "@" + System.identityHashCode(this);
//		return ToStringBuilder.reflectionToString(this, ToStringStyle.SHORT_PREFIX_STYLE);
	}
	
}
