/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2008/09/17
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.model.attribute;

import org.jiemamy.model.attribute.constraint.ColumnCheckConstraint;
import org.jiemamy.model.attribute.constraint.NotNullConstraint;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.datatype.DataType;
import org.jiemamy.model.datatype.DataTypeHolder;

/**
 * リレーショナルデータベースにおける「カラム」を表すモデルインターフェイス。
 * 
 * @since 0.2
 * @author daisuke
 */
public interface ColumnModel extends AttributeModel, DataTypeHolder<DataType> {
	
	/**
	 * チェック制約を取得する。
	 * 
	 * @return　チェック制約. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	ColumnCheckConstraint getCheckConstraint();
	
	/**
	 * 型記述子を取得する。
	 * 
	 * @return 型記述子. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	DataType getDataType();
	
	/**
	 * デフォルト値を取得する。
	 * 
	 * @return デフォルト値. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	String getDefaultValue();
	
	/**
	 * NOT　NULL制約を取得する。
	 * 
	 * @return　NOT　NULL制約. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	NotNullConstraint getNotNullConstraint();
	
	/**
	 * 主キー制約を取得する。
	 * 
	 * @return　主キー制約. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	PrimaryKey getPrimaryKey();
	
	/**
	 * 一意キー制約を取得する。
	 * 
	 * @return　一意キー制約. 未設定の場合は{@code null}
	 * @since 0.2
	 */
	UniqueKey getUniqueKey();
	
	/**
	 * カラムが主キー制約を構成するカラムであるかどうか調べる。
	 * 
	 * @return 主キー制約構成カラムである場合は{@code true}、そうでない場合は{@code false}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @throws UnsupportedOperationException 実装がこのメソッドをサポートしない場合
	 */
	boolean checkPrimaryKeyColumn();
	
	/**
	 * チェック制約を設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param checkConstraint チェック制約
	 * @since 0.2
	 */
	void setCheckConstraint(ColumnCheckConstraint checkConstraint);
	
	/**
	 * 型記述子を設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与えるが、モデルとしては無効となる。</p>
	 * 
	 * @param dataType 型記述子
	 * @since 0.2
	 */
	void setDataType(DataType dataType);
	
	/**
	 * デフォルト値を設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param defaultValue デフォルト値
	 * @since 0.2
	 */
	void setDefaultValue(String defaultValue);
	
	/**
	 * NOT　NULL制約を設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param notNullConstraint NOT　NULL制約
	 * @since 0.2
	 */
	void setNotNullConstraint(NotNullConstraint notNullConstraint);
	
	/**
	 * 主キー制約を設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param primaryKey 主キー制約
	 * @since 0.2
	 */
	void setPrimaryKey(PrimaryKey primaryKey);
	
	/**
	 * 一意キー制約を設定する。
	 * 
	 * <p>未設定とする場合は{@code null}を与える。</p>
	 * 
	 * @param uniqueKey 一意キー制約
	 * @since 0.2
	 */
	void setUniqueKey(UniqueKey uniqueKey);
	
}
