/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/29
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils;

import javax.xml.namespace.QName;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.xml.JiemamyQName;

/**
 * DOMのユーティリティクラス。
 * 
 * @author daisuke
 */
public final class DomUtil {
	
	/**
	 * XML要素が、指定された完全修飾名を持つ要素かどうかを調べる。
	 * 
	 * @param element 検査対象要素
	 * @param qName 完全修飾名
	 * @return 要素が指定された完全修飾名を持っていれば{@code true}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static boolean isEqualsQName(Element element, JiemamyQName qName) {
		Validate.notNull(element);
		Validate.notNull(qName);
		
		return isEqualsQName(element, qName.getQName());
	}
	
	/**
	 * XML要素が、指定された完全修飾名を持つ要素かどうかを調べる。
	 * 
	 * @param element 検査対象要素
	 * @param qName 完全修飾名
	 * @return 要素が指定された完全修飾名を持っている場合は{@code true}、そうでない場合は{@code false}. 
	 *   引数に{@code null}を与えた場合は常に{@code false}
	 */
	public static boolean isEqualsQName(Element element, QName qName) {
		if (element == null || qName == null) {
			return false;
		}
		return qName.getNamespaceURI().equals(element.getNamespaceURI())
				&& qName.getLocalPart().equals(element.getTagName());
	}
	
	/**
	 * 新しい要素を生成し、指定された要素の子として追加して返す。
	 * 
	 * @param parentElement 親要素
	 * @param qName 完全修飾要素名
	 * @return 生成された要素
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static Element newChild(Element parentElement, JiemamyQName qName) {
		Validate.notNull(parentElement);
		Validate.notNull(qName);
		return newChild(parentElement, qName.getQName());
	}
	
	/**
	 * 新しい要素を生成し、指定された要素の子として追加して返す。
	 * 新しい要素の内容として、与えた {@link Enum#name()} をcharactersとして追加する。
	 * 
	 * @param parentElement 親要素
	 * @param qName 完全修飾要素名
	 * @param e 設定値
	 * @return 生成された要素
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static Element newChild(Element parentElement, JiemamyQName qName, Enum<?> e) {
		Validate.notNull(parentElement);
		Validate.notNull(qName);
		return newChild(parentElement, qName.getQName(), e);
	}
	
	/**
	 * 新しい要素を生成し、指定された要素の子として追加して返す。
	 * 新しい要素の内容として、与えたtextをcharactersとして追加する。
	 * 
	 * @param parentElement 親要素
	 * @param qName 完全修飾要素名
	 * @param text 設定値
	 * @return 生成された要素
	 * @throws IllegalArgumentException 引数parentElement, qNameに{@code null}を与えた場合
	 */
	public static Element newChild(Element parentElement, JiemamyQName qName, String text) {
		Validate.notNull(parentElement);
		Validate.notNull(qName);
		return newChild(parentElement, qName.getQName(), text);
	}
	
	/**
	 * 新しい要素を生成し、指定された要素の子として追加して返す。
	 * 
	 * @param parentElement 親要素
	 * @param qName 完全修飾要素名
	 * @return 生成された要素
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static Element newChild(Element parentElement, QName qName) {
		Validate.notNull(parentElement);
		Validate.notNull(qName);
		Element childElement = newElement(parentElement.getOwnerDocument(), qName);
		parentElement.appendChild(childElement);
		return childElement;
	}
	
	/**
	 * 新しい要素を生成し、指定された要素の子として追加して返す。
	 * 新しい要素の内容として、与えた {@link Enum#name()} をcharactersとして追加する。
	 * 
	 * <p>eが{@code null}だった場合、要素は追加せず、{@code null}を返す。</p>
	 * 
	 * @param parentElement 親要素
	 * @param qName 完全修飾要素名
	 * @param e 設定値
	 * @return 生成された要素
	 * @throws IllegalArgumentException 引数parentElement, qNameに{@code null}を与えた場合
	 */
	public static Element newChild(Element parentElement, QName qName, Enum<?> e) {
		Validate.notNull(parentElement);
		Validate.notNull(qName);
		Element childElement = null;
		if (e != null) {
			childElement = newChild(parentElement, qName, e.name());
		}
		return childElement;
	}
	
	/**
	 * 新しい要素を生成し、指定された要素の子として追加して返す。
	 * 新しい要素の内容として、与えたtextをcharactersとして追加する。
	 * 
	 * <p>testが{@code null}だった場合、要素は追加せず{@code null}を返す。</p>
	 * 
	 * @param parentElement 親要素
	 * @param qName 完全修飾要素名
	 * @param text 設定値 may be null
	 * @return 生成された要素
	 * @throws IllegalArgumentException 引数parentElement, qNameに{@code null}を与えた場合
	 */
	public static Element newChild(Element parentElement, QName qName, String text) {
		Validate.notNull(parentElement);
		Validate.notNull(qName);
		Element childElement = null;
		if (text != null) {
			childElement = newChild(parentElement, qName);
			setText(childElement, text);
		}
		return childElement;
	}
	
	/**
	 * 新しい要素を生成して返す。
	 * 
	 * @param document DOMドキュメント
	 * @param qName 完全修飾要素名
	 * @return 生成された要素
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static Element newElement(Document document, JiemamyQName qName) {
		Validate.notNull(document);
		Validate.notNull(qName);
		return newElement(document, qName.getQName());
	}
	
	/**
	 * 新しい要素を生成して返す。
	 * 
	 * @param document DOMドキュメント
	 * @param qName 完全修飾要素名
	 * @return 生成された要素
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static Element newElement(Document document, QName qName) {
		Validate.notNull(document);
		Validate.notNull(qName);
		
		String prefix = StringUtils.isEmpty(qName.getPrefix()) ? "" : qName.getPrefix() + ":";
		return document.createElementNS(qName.getNamespaceURI(), prefix + qName.getLocalPart());
	}
	
	/**
	 * 要素の属性に値を設定する。
	 * 
	 * @param domElement 要素
	 * @param node 属性指定
	 * @param value 値
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static void setAttribute(Element domElement, JiemamyQName node, String value) {
		Validate.notNull(domElement);
		Validate.notNull(node);
		Validate.notNull(value);
		
		domElement.setAttribute(node.getQNameString(), value);
	}
	
	/**
	 * 与えた要素のcharacters子ノードとして文字列をappendする。
	 * 
	 * <p>与えた文字列が{@code null}だった場合、何もしない。</p>
	 * 
	 * @param element 対象要素
	 * @param text 文字列
	 * @throws IllegalArgumentException 引数elementに{@code null}を与えた場合
	 */
	public static void setText(Element element, String text) {
		Validate.notNull(element);
		if (text != null) {
			element.appendChild(element.getOwnerDocument().createTextNode(text));
		}
	}
	
	private DomUtil() {
		throw new UnexpectedConditionError("不到達ポイント");
	}
	
}
