/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/04/21
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.commons.lang.Validate;
import org.w3c.dom.Attr;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.jiemamy.Jiemamy;
import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.xml.JiemamyNamespaceContext;

/**
 * Jiemamy上で使用する{@link XPath}をタイプセーフに扱うためのラッパークラス。
 * 
 * @author daisuke
 */
public class SmartXpath {
	
	/** XPathインスタンス */
	public final XPath xpath = XPathFactory.newInstance().newXPath();
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param jiemamy コンテキストインスタンス
	 */
	public SmartXpath(Jiemamy jiemamy) {
		xpath.setNamespaceContext(new JiemamyNamespaceContext(jiemamy.getNamespaces()));
	}
	
	/**
	 * XPath式で特定される{@link Attr}を取得する。
	 * 
	 * @param root 対象ノード
	 * @param strExpr XPath式
	 * @return 取得した {@link Attr}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public Attr getAttr(Node root, String strExpr) {
		Validate.notNull(root);
		Validate.notNull(strExpr);
		
		try {
			return (Attr) xpath.evaluate(strExpr, root, XPathConstants.NODE);
		} catch (XPathExpressionException e) {
			throw new UnexpectedConditionError("XPath expression error: " + strExpr, e);
		}
	}
	
	/**
	 * XPath式で特定される{@link Node}を取得する。
	 * 
	 * @param root 対象ノード
	 * @param strExpr XPath式
	 * @return 取得した {@link Node}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public Node getNode(Node root, String strExpr) {
		Validate.notNull(root);
		Validate.notNull(strExpr);
		
		try {
			return (Node) xpath.evaluate(strExpr, root, XPathConstants.NODE);
		} catch (XPathExpressionException e) {
			throw new UnexpectedConditionError("XPath expression error: " + strExpr, e);
		}
	}
	
	/**
	 * XPath式で特定される{@link NodeList}を取得する。
	 * 
	 * @param root 対象ノード
	 * @param strExpr XPath式
	 * @return 取得した {@link NodeList}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public NodeList getNodes(Node root, String strExpr) {
		Validate.notNull(root);
		Validate.notNull(strExpr);
		
		try {
			return (NodeList) xpath.evaluate(strExpr, root, XPathConstants.NODESET);
		} catch (XPathExpressionException e) {
			throw new UnexpectedConditionError("XPath expression error: " + strExpr, e);
		}
	}
	
	/**
	 * XPath式で特定されるノードの内容文字列を取得する。
	 * 
	 *<p> ノードが見つからなかった場合は{@code null}を返す。ノードはあるが空だった場合は空文字列を返す。</p>
	 * 
	 * @param root 対象ノード
	 * @param strExpr XPath式
	 * @return 取得した文字列
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public String getTextContent(Node root, String strExpr) {
		Validate.notNull(root);
		Validate.notNull(strExpr);
		
		try {
			Node node = (Node) xpath.evaluate(strExpr, root, XPathConstants.NODE);
			if (node == null) {
				return null;
			}
			return node.getTextContent();
		} catch (XPathExpressionException e) {
			throw new UnexpectedConditionError("XPath expression error: " + strExpr, e);
		}
	}
}
