/*
 * Copyright 2009 Jiemamy Project and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils.enhancer.helper;

import static org.jiemamy.utils.enhancer.helper.EnhanceManipulator.createAdviceTableField;
import static org.jiemamy.utils.enhancer.helper.EnhanceManipulator.replaceToPointcut;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javassist.CannotCompileException;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.NotFoundException;
import javassist.expr.ExprEditor;
import javassist.expr.NewExpr;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.jiemamy.utils.enhancer.EnhanceException;
import org.jiemamy.utils.enhancer.InvocationHandler;

/**
 * ファクトリに含まれるインスタンス生成式を拡張する。
 * @version $Date: 2009-01-21 20:37:40 +0900 (水, 21  1月 2009) $
 * @author Suguru ARAKAWA (Gluegent, Inc.)
 */
public class NewInstanceEnhancer extends ExprEditor {
	
	private static final Logger LOG = LoggerFactory.getLogger(NewInstanceEnhancer.class);
	

	/**
	 * 指定のファクトリクラスに、アスペクト用のフックを織り込む。
	 * <p>
	 * 実際に織り込まれるアスペクトの一覧は、{@code enhanceList}と
	 * プロダクトの拡張方法を元に計算される。
	 * このメソッドでは、{@code enhanceList}に規定される拡張の一覧を
	 * 実際に適用するためのフックのみを織り込む。
	 * </p>
	 * <p>
	 * {@code productAspects.base IN productsToBeEnhanced.base}
	 * </p>
	 * @param target アスペクトを埋め込む先のファクトリクラス
	 * @param enhanceManager 拡張の定義
	 * @param productsToBeEnhanced
	 *      拡張されるべきプロダクトクラスの一覧 ({@code base -> toBeEnhanced})
	 * @param allProductAspects
	 *      それぞれのプロダクトクラスに対するメソッドアスペクトの一覧
	 *      ({@code base -> aspect for each method})
	 * @return
	 *      ファクトリに実際に埋め込まれるべきアスペクトの一覧
	 * @throws EnhanceException 拡張に失敗した場合
	 */
	public static AspectList<CtConstructor> enhance(CtClass target, EnhanceManager enhanceManager,
			Map<? extends CtClass, ? extends CtClass> productsToBeEnhanced,
			Map<? extends CtClass, AspectList<CtMethod>> allProductAspects) throws EnhanceException {
		
		if (target == null) {
			throw new NullPointerException("target"); //$NON-NLS-1$
		}
		if (enhanceManager == null) {
			throw new NullPointerException("enhanceManager"); //$NON-NLS-1$
		}
		if (productsToBeEnhanced == null) {
			throw new NullPointerException("productsToBeEnhanced"); //$NON-NLS-1$
		}
		if (allProductAspects == null) {
			throw new NullPointerException("allProductAspects"); //$NON-NLS-1$
		}
		
		try {
			NewInstanceEnhancer enhancer =
					new NewInstanceEnhancer(target, enhanceManager, productsToBeEnhanced, allProductAspects);
			target.instrument(enhancer);
			if (enhancer.aspects.isEmpty()) {
				assert enhancer.holder == null;
				return null;
			} else {
				assert enhancer.holder != null;
				return new AspectList<CtConstructor>(enhancer.holder, enhancer.aspects);
			}
		} catch (CannotCompileException e) {
			throw new EnhanceException(MessageFormat.format("Cannot compile {0}", target.getName()), e);
		}
	}
	

	private final CtClass target;
	
	private final EnhanceManager enhanceManager;
	
	private final Map<? extends CtClass, ? extends CtClass> productEnhanceMap;
	
	private final Map<? extends CtClass, AspectList<CtMethod>> allProductAspects;
	
	private final List<Aspect<CtConstructor>> aspects;
	
	private int enhanceIndex;
	
	private CtField holder;
	

	private NewInstanceEnhancer(CtClass target, EnhanceManager enhanceManager,
			Map<? extends CtClass, ? extends CtClass> productsToBeEnhanced,
			Map<? extends CtClass, AspectList<CtMethod>> allProductAspects) {
		super();
		assert target != null;
		assert enhanceManager != null;
		assert productsToBeEnhanced != null;
		assert allProductAspects != null;
		this.target = target;
		this.enhanceManager = enhanceManager;
		productEnhanceMap = productsToBeEnhanced;
		this.allProductAspects = allProductAspects;
		aspects = new ArrayList<Aspect<CtConstructor>>();
		enhanceIndex = 0;
		holder = null;
	}
	
	/**
	 * それぞれのインスタンス生成式を必要ならば書き換える。
	 */
	@Override
	public void edit(NewExpr expr) throws CannotCompileException {
		LOG.trace("Pointcut candidate: new {}", expr.getClassName());
		try {
			CtConstructor constructor = expr.getConstructor();
			if (enhanceManager.isLegalJoinpoint(constructor) == false) {
				return;
			}
			CtClass declaring = constructor.getDeclaringClass();
			List<InvocationHandler> handlers = enhanceManager.findApplicableHandlers(declaring, constructor);
			if (handlers.isEmpty() && productEnhanceMap.containsKey(declaring) == false
					&& allProductAspects.containsKey(declaring) == false) {
				return;
			}
			if (enhanceIndex == 0) {
				holder = createAdviceTableField(target);
			}
			
			LOG.debug("Rewrite new[{}]: {} (at {}{}:{})", new Object[] {
				enhanceIndex,
				declaring.getName(),
				expr.where().getName(),
				expr.where().getSignature(),
				expr.getLineNumber()
			});
			
			replaceToPointcut(expr, holder, enhanceIndex);
			
			if (productEnhanceMap.containsKey(declaring)) {
				CtClass actual = productEnhanceMap.get(declaring);
				String descriptor = constructor.getSignature();
				CtConstructor actualCtor = actual.getConstructor(descriptor);
				aspects.add(new Aspect<CtConstructor>(constructor, actualCtor, handlers));
			} else {
				assert allProductAspects.containsKey(declaring) == false;
				aspects.add(new Aspect<CtConstructor>(constructor, constructor, handlers));
			}
			
			enhanceIndex++;
		} catch (EnhanceException e) {
			throw new CannotCompileException(e);
		} catch (NotFoundException e) {
			throw new CannotCompileException(e);
		}
	}
}
