/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/19
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils.model;

import java.util.Collection;
import java.util.List;
import java.util.UUID;

import org.apache.commons.lang.Validate;

import org.jiemamy.dialect.Dialect;
import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.constraint.ColumnCheckConstraint;
import org.jiemamy.model.attribute.constraint.Deferrability;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.NotNullConstraint;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.dataset.RecordModel;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.index.IndexModel;
import org.jiemamy.utils.CollectionsUtil;

/**
 * {@link RootModel}のユーティリティクラス。
 * 
 * @author daisuke
 */
public final class RootModelUtil {
	
	/**
	 * {@link RootModel}をトラバースし、指定したIDを持つモデルを取得する。
	 * 
	 * @param rootModel 対象RootModel
	 * @param refId ID
	 * @return 見つかったモデル。見つからなかった場合は{@code null}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static JiemamyElement findJiemamyElement(RootModel rootModel, UUID refId) {
		Validate.notNull(rootModel);
		Validate.notNull(refId);
		JiemamyElement result = null;
		for (DomainModel domainModel : rootModel.getDomains()) {
			if (domainModel.getId().equals(refId)) {
				result = domainModel;
				break;
			}
			NotNullConstraint notNull = domainModel.getNotNullConstraint();
			if (notNull != null && refId.equals(notNull.getId())) {
				result = notNull;
				break;
			}
			ColumnCheckConstraint check = domainModel.getCheckConstraint();
			if (check != null && refId.equals(check.getId())) {
				result = check;
				break;
			}
		}
		if (result != null) {
			return result;
		}
		
		for (EntityModel entityModel : rootModel.getEntities()) {
			if (refId.equals(entityModel.getId())) {
				result = entityModel;
				break;
			}
			if (entityModel instanceof TableModel) {
				TableModel tableModel = (TableModel) entityModel;
				for (AttributeModel attributeModel : tableModel.getAttributes()) {
					if (refId.equals(attributeModel.getId())) {
						result = attributeModel;
						break;
					}
					if (attributeModel instanceof ColumnModel) {
						ColumnModel columnModel = (ColumnModel) attributeModel;
						NotNullConstraint notNull = columnModel.getNotNullConstraint();
						if (notNull != null && refId.equals(notNull.getId())) {
							result = notNull;
							break;
						}
						UniqueKey uniqueKey = columnModel.getUniqueKey();
						if (uniqueKey != null && refId.equals(uniqueKey.getId())) {
							result = uniqueKey;
							break;
						}
						PrimaryKey primaryKey = columnModel.getPrimaryKey();
						if (primaryKey != null && refId.equals(primaryKey.getId())) {
							result = primaryKey;
							break;
						}
						ColumnCheckConstraint check = columnModel.getCheckConstraint();
						if (check != null && refId.equals(check.getId())) {
							result = check;
							break;
						}
					}
					if (attributeModel instanceof ForeignKey) {
						ForeignKey foreignKey = (ForeignKey) attributeModel;
						Deferrability deferrability = foreignKey.getDeferrability();
						if (deferrability != null && refId.equals(deferrability.getId())) {
							result = deferrability;
							break;
						}
					}
				}
				if (result != null) {
					return result;
				}
				
				for (IndexModel indexModel : tableModel.getIndexes()) {
					if (refId.equals(indexModel.getId())) {
						result = indexModel;
					}
				}
				if (result != null) {
					return result;
				}
			}
		}
		
		for (DataSetModel dataSetModel : rootModel.getDataSets()) {
			if (refId.equals(dataSetModel.getId())) {
				result = dataSetModel;
				break;
			}
			for (List<RecordModel> records : dataSetModel.getRecords().values()) {
				for (RecordModel recordModel : records) {
					if (refId.equals(recordModel.getId())) {
						result = recordModel;
						break;
					}
				}
			}
		}
		
		return result;
	}
	
	/**
	 * この {@link RootModel} 内に存在する全ての外部キーの集合を取得する。
	 * 
	 * @param rootModel 検索対象の{@link RootModel}
	 * @return 外部キーの集合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static Collection<ForeignKey> getAllForeignKeys(RootModel rootModel) {
		Validate.notNull(rootModel);
		List<ForeignKey> result = CollectionsUtil.newArrayList();
		for (TableModel tableModel : rootModel.findEntities(TableModel.class)) {
			result.addAll(tableModel.findAttributes(ForeignKey.class));
		}
		return result;
	}
	
	/**
	 * SQL方言を設定する。
	 * 
	 * @param rootModel 操作対象{@link RootModel}
	 * @param dialectClass SQL方言クラス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static void setDialect(RootModel rootModel, Class<? extends Dialect> dialectClass) {
		Validate.notNull(rootModel);
		Validate.notNull(dialectClass);
		rootModel.setDialectClassName(dialectClass.getName());
	}
	
	/**
	 * SQL方言を設定する。
	 * 
	 * @param rootModel 操作対象{@link RootModel}
	 * @param dialect SQL方言
	 */
	public static void setDialect(RootModel rootModel, Dialect dialect) {
		Validate.notNull(rootModel);
		Validate.notNull(dialect);
		setDialect(rootModel, dialect.getClass());
	}
	
	private RootModelUtil() {
		throw new UnexpectedConditionError("不到達ポイント");
	}
}
