/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/26
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.validator.impl;

import java.util.Collection;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.constraint.LocalKeyConstraint;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.validator.AbstractProblem;
import org.jiemamy.validator.AbstractValidator;
import org.jiemamy.validator.Problem;

/**
 * テーブルに存在するカラムの数のバリデータ。
 * 
 * <ul>
 *   <li>テーブルにカラムは1つ以上必要である。</li>
 *   <li>カラムには、データ型を設定する必要がある。</li>
 *   <li>カラムに対して設定されたキー制約は、キーカラムのサイズが1でなければならない。</li>
 * </ul>
 * 
 * @author daisuke
 */
public class ColumnValidator extends AbstractValidator {
	
	@Override
	public Collection<Problem> validate(RootModel rootModel) {
		Collection<Problem> result = super.validate(rootModel);
		Collection<TableModel> tableModels = rootModel.findEntities(TableModel.class);
		for (TableModel tableModel : tableModels) {
			int index = 0;
			for (ColumnModel columnModel : tableModel.findColumns()) {
				if (StringUtils.isEmpty(columnModel.getName())) {
					result.add(new EmptyColumnNameProblem(tableModel, index));
				}
				if (columnModel.getDataType() == null) {
					result.add(new EmptyDataTypeProblem(tableModel, columnModel));
				}
				
				UniqueKey uniqueKey = columnModel.getUniqueKey();
				if (uniqueKey != null && uniqueKey.getKeyColumns().size() != 1) {
					result.add(new MultiKeyColumnProblem(columnModel, uniqueKey));
				}
				PrimaryKey primaryKey = columnModel.getPrimaryKey();
				if (primaryKey != null && primaryKey.getKeyColumns().size() != 1) {
					result.add(new MultiKeyColumnProblem(columnModel, primaryKey));
				}
				index++;
			}
		}
		return result;
	}
	

	static class EmptyColumnNameProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param tableModel 無名カラムが含まれるテーブル
		 * @param index カラムのインデックス
		 */
		public EmptyColumnNameProblem(TableModel tableModel, int index) {
			super("E0040");
			registerMessage(Locale.JAPAN, "テーブル{0}の{1}番目のカラムに名前がありません");
			setArguments(new Object[] {
				tableModel.getName(),
				index
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
	
	static class EmptyDataTypeProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param tableModel データ型無指定のカラムが含まれるテーブル
		 * @param columnModel データ型無指定のカラム
		 */
		public EmptyDataTypeProblem(TableModel tableModel, ColumnModel columnModel) {
			super("E0050");
			registerMessage(Locale.JAPAN, "テーブル{0}のカラム{1}にデータ型が設定されていません");
			setArguments(new Object[] {
				tableModel.getName(),
				columnModel.getName()
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
	
	static class MultiKeyColumnProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param columnModel 複合キーが設定されたカラム
		 * @param localKeyConstraint 複合キー
		 */
		public MultiKeyColumnProblem(ColumnModel columnModel, LocalKeyConstraint localKeyConstraint) {
			super("E0060");
			registerMessage(Locale.JAPAN, "カラム{0}に対するキー制約({1})において、キーカラムの構成カラムが無いまたは複数で構成されています");
			setArguments(new Object[] {
				columnModel.getName(),
				StringUtils.isEmpty(localKeyConstraint.getName()) ? localKeyConstraint.getId().toString()
						: localKeyConstraint.getName()
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
}
