/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/26
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.validator.impl;

import java.util.Collection;
import java.util.Locale;

import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.validator.AbstractProblem;
import org.jiemamy.validator.AbstractValidator;
import org.jiemamy.validator.Problem;

/**
 * {@link PrimaryKey}の数を調べるバリデータ。
 * 
 * <ul>
 *   <li>1つのテーブルには複数の{@link PrimaryKey}が含まれていてはならない。</li>
 *   <li>NOTICE: テーブルに{@link PrimaryKey}が1つも存在しなかった場合。</p>
 * </ul>
 * 
 * @author daisuke
 */
public class PrimaryKeyValidator extends AbstractValidator {
	
	@Override
	public Collection<Problem> validate(RootModel rootModel) {
		Collection<Problem> result = super.validate(rootModel);
		Collection<TableModel> tableModels = rootModel.findEntities(TableModel.class);
		for (TableModel tableModel : tableModels) {
			int size = tableModel.findAttributes(PrimaryKey.class, true).size();
			if (size == 0) {
				result.add(new NoPrimaryKeyProblem(tableModel));
			} else if (size > 1) {
				result.add(new MultiplePrimaryKeyProblem(tableModel));
			}
		}
		return result;
	}
	

	static class MultiplePrimaryKeyProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param tableModel 複数の主キーが設定されたテーブル
		 */
		public MultiplePrimaryKeyProblem(TableModel tableModel) {
			super("E0140");
			registerMessage(Locale.JAPAN, "テーブル{0}において主キー制約が複数存在しています");
			setArguments(new Object[] {
				tableModel.getName()
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
	
	static class NoPrimaryKeyProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param tableModel 主キーが存在しないテーブル
		 */
		public NoPrimaryKeyProblem(TableModel tableModel) {
			super("N0010");
			registerMessage(Locale.JAPAN, "テーブル{0}において主キー制約が存在しません");
			setArguments(new Object[] {
				tableModel.getName()
			});
		}
		
		public Severity getSeverity() {
			return Severity.NOTICE;
		}
		
	}
}
